from collections.abc import Mapping
from pathlib import Path
from typing import TYPE_CHECKING, Literal, TypeVar

import jax
import jax.numpy as jnp
import numpy as np
import numpyro

from astropy.io import fits
from jax.experimental.sparse import BCOO
from numpyro import handlers

from ..model.abc import SpectralModel
from ..util.online_storage import table_manager
from . import Instrument, ObsConfiguration, Observation

K = TypeVar("K")
V = TypeVar("V")

if TYPE_CHECKING:
    from ..data import ObsConfiguration
    from ..model.abc import SpectralModel


def load_example_pha(
    source: Literal["NGC7793_ULX4_PN", "NGC7793_ULX4_ALL"],
) -> (Observation, list[Observation] | dict[str, Observation]):
    """
    Load some example observations from the package data.

    Parameters:
        source: The source to be loaded. Can be either "NGC7793_ULX4_PN" or "NGC7793_ULX4_ALL".
    """

    if source == "NGC7793_ULX4_PN":
        return Observation.from_pha_file(
            table_manager.fetch("example_data/NGC7793_ULX4/PN_spectrum_grp20.fits"),
            bkg_path=table_manager.fetch("example_data/NGC7793_ULX4/PNbackground_spectrum.fits"),
        )

    elif source == "NGC7793_ULX4_ALL":
        return {
            "PN": Observation.from_pha_file(
                table_manager.fetch("example_data/NGC7793_ULX4/PN_spectrum_grp20.fits"),
                bkg_path=table_manager.fetch(
                    "example_data/NGC7793_ULX4/PNbackground_spectrum.fits"
                ),
            ),
            "MOS1": Observation.from_pha_file(
                table_manager.fetch("example_data/NGC7793_ULX4/MOS1_spectrum_grp.fits"),
                bkg_path=table_manager.fetch(
                    "example_data/NGC7793_ULX4/MOS1background_spectrum.fits"
                ),
            ),
            "MOS2": Observation.from_pha_file(
                table_manager.fetch("example_data/NGC7793_ULX4/MOS2_spectrum_grp.fits"),
                bkg_path=table_manager.fetch(
                    "example_data/NGC7793_ULX4/MOS2background_spectrum.fits"
                ),
            ),
        }

    else:
        raise ValueError(f"{source} not recognized.")


def load_example_instruments(source: Literal["NGC7793_ULX4_PN", "NGC7793_ULX4_ALL"]):
    """
    Load some example instruments from the package data.

    Parameters:
        source: The source to be loaded. Can be either "NGC7793_ULX4_PN" or "NGC7793_ULX4_ALL".

    """
    if source == "NGC7793_ULX4_PN":
        return Instrument.from_ogip_file(
            table_manager.fetch("example_data/NGC7793_ULX4/PN.rmf"),
            table_manager.fetch("example_data/NGC7793_ULX4/PN.arf"),
        )

    elif source == "NGC7793_ULX4_ALL":
        return {
            "PN": Instrument.from_ogip_file(
                table_manager.fetch("example_data/NGC7793_ULX4/PN.rmf"),
                table_manager.fetch("example_data/NGC7793_ULX4/PN.arf"),
            ),
            "MOS1": Instrument.from_ogip_file(
                table_manager.fetch("example_data/NGC7793_ULX4/MOS1.rmf"),
                table_manager.fetch("example_data/NGC7793_ULX4/MOS1.arf"),
            ),
            "MOS2": Instrument.from_ogip_file(
                table_manager.fetch("example_data/NGC7793_ULX4/MOS2.rmf"),
                table_manager.fetch("example_data/NGC7793_ULX4/MOS2.arf"),
            ),
        }

    else:
        raise ValueError(f"{source} not recognized.")


def load_example_obsconf(source: Literal["NGC7793_ULX4_PN", "NGC7793_ULX4_ALL"]):
    """
    Load some example ObsConfigurations.

    Parameters:
        source: The source to be loaded. Can be either "NGC7793_ULX4_PN" or "NGC7793_ULX4_ALL".
    """

    if source in "NGC7793_ULX4_PN":
        instrument = load_example_instruments(source)
        observation = load_example_pha(source)

        return ObsConfiguration.from_instrument(
            instrument, observation, low_energy=0.5, high_energy=8.0
        )

    elif source == "NGC7793_ULX4_ALL":
        instruments_dict = load_example_instruments(source)
        observations_dict = load_example_pha(source)

        return {
            key: ObsConfiguration.from_instrument(
                instruments_dict[key], observations_dict[key], low_energy=0.5, high_energy=8.0
            )
            for key in instruments_dict.keys()
        }

    else:
        raise ValueError(f"{source} not recognized.")


def forward_model_with_multiple_inputs(
    model: "SpectralModel",
    parameters,
    obs_configuration: "ObsConfiguration",
    sparse=False,
):
    energies = np.asarray(obs_configuration.in_energies)
    parameter_dims = next(iter(parameters.values())).shape

    def flux_func(p):
        return model.photon_flux(p, *energies)

    for _ in parameter_dims:
        flux_func = jax.vmap(flux_func)

    flux_func = jax.jit(flux_func)

    if sparse:
        # folding.transfer_matrix.data.density > 0.015 is a good criterion to consider sparsify
        transfer_matrix = BCOO.from_scipy_sparse(
            obs_configuration.transfer_matrix.data.to_scipy_sparse().tocsr()
        )

    else:
        transfer_matrix = np.asarray(obs_configuration.transfer_matrix.data.todense())

    expected_counts = jnp.matvec(transfer_matrix, flux_func(parameters))

    # The result is clipped at 1e-6 to avoid 0 round-off and diverging likelihoods
    return jnp.clip(expected_counts, a_min=1e-6)


def fakeit_for_multiple_parameters(
    obsconfs: ObsConfiguration | list[ObsConfiguration],
    model: SpectralModel,
    parameters: Mapping[K, V],
    rng_key: int = 0,
    apply_stat: bool = True,
    sparsify_matrix: bool = False,
):
    """
    Convenience function to simulate multiple spectra from a given model and a set of parameters.
    This is supposed to be somewhat optimized and can handle multiple parameters at once without blowing
    up the memory. The parameters should be passed as a dictionary with the parameter name as the key and
    the parameter values as the values, the value can be a scalar or a nd-array.

    # Example:

    ``` python
    from jaxspec.data.util import fakeit_for_multiple_parameters
    from numpy.random import default_rng

    rng = default_rng(42)
    size = (10, 30)

    parameters = {
        "tbabs_1_nh": rng.uniform(0.1, 0.4, size=size),
        "powerlaw_1_alpha": rng.uniform(1, 3, size=size),
        "powerlaw_1_norm": rng.exponential(10 ** (-0.5), size=size),
        "blackbodyrad_1_kT": rng.uniform(0.1, 3.0, size=size),
        "blackbodyrad_1_norm": rng.exponential(10 ** (-3), size=size)
    }

    spectra = fakeit_for_multiple_parameters(obsconf, model, parameters)
    ```

    Parameters:
        obsconfs: The observational setup(s).
        model: The model to use.
        parameters: The parameters of the model.
        rng_key: The random number generator seed.
        apply_stat: Whether to apply Poisson statistic on the folded spectra or not.
        sparsify_matrix: Whether to sparsify the matrix or not.
    """

    obsconf_list = [obsconfs] if isinstance(obsconfs, ObsConfiguration) else obsconfs
    fakeits = []

    for i, obsconf in enumerate(obsconf_list):
        countrate = forward_model_with_multiple_inputs(
            model, parameters, obsconf, sparse=sparsify_matrix
        )

        if apply_stat:
            with handlers.seed(rng_seed=rng_key):
                spectrum = numpyro.sample(
                    f"likelihood_obs_{i}",
                    numpyro.distributions.Poisson(countrate),
                )

        else:
            spectrum = countrate

        fakeits.append(spectrum)

    return fakeits[0] if len(fakeits) == 1 else fakeits


def data_path_finder(
    pha_path: str, require_arf: bool = True, require_rmf: bool = True, require_bkg: bool = False
) -> tuple[str | None, str | None, str | None]:
    """
    Function which tries its best to find the ARF, RMF and BKG files associated with a given PHA file.

    Parameters:
        pha_path: The PHA file path.
        require_arf: Whether to raise an error if the ARF file is not found.
        require_rmf: Whether to raise an error if the RMF file is not found.
        require_bkg: Whether to raise an error if the BKG file is not found.

    Returns:
        arf_path: The ARF file path.
        rmf_path: The RMF file path.
        bkg_path: The BKG file path.
    """

    def find_path(file_name: str, directory: str, raise_err: bool = True) -> str | None:
        if raise_err:
            if file_name.lower() != "none" and file_name != "":
                return find_file_or_compressed_in_dir(file_name, directory, raise_err)

        return None

    header = fits.getheader(pha_path, "SPECTRUM")
    directory = str(Path(pha_path).parent)

    arf_path = find_path(header.get("ANCRFILE", "none"), directory, require_arf)
    rmf_path = find_path(header.get("RESPFILE", "none"), directory, require_rmf)
    bkg_path = find_path(header.get("BACKFILE", "none"), directory, require_bkg)

    return arf_path, rmf_path, bkg_path


def find_file_or_compressed_in_dir(path: str | Path, directory: str | Path, raise_err: bool) -> str:
    """
    Try to find a file or its .gz compressed version in a given directory and return
    the full path of the file.
    """
    path = Path(path) if isinstance(path, str) else path
    directory = Path(directory) if isinstance(directory, str) else directory

    if directory.joinpath(path).exists():
        return str(directory.joinpath(path))

    matching_files = list(directory.glob(str(path) + "*"))

    if matching_files:
        file = matching_files[0]
        if file.suffix == ".gz":
            return str(file)

    elif raise_err:
        raise FileNotFoundError(f"Can't find {path}(.gz) in {directory}.")



import astropy.units as u
import matplotlib.pyplot as plt
import pytest

from jaxspec.analysis.compare import plot_corner_comparison


def test_plot_ppc(get_result_list):
    for name, result in zip(*get_result_list):
        result.plot_ppc(plot_components=True, plot_background=False)


def test_plot_scales(get_result_list):
    name, result = next(zip(*get_result_list))
    for case in ["linear", "semilogx", "semilogy", "loglog"]:
        result.plot_ppc(plot_components=True, plot_background=False, scale=case)


def test_plot_ppc_components(get_result_list):
    for name, result in zip(*get_result_list):
        result.plot_ppc(n_sigmas=2)


def test_plot_ppc_units(get_result_list):
    for name, result in zip(*get_result_list):
        for x_unit in ["angstrom", "keV", "Hz", "nm"]:
            result.plot_ppc(x_unit=x_unit)


def test_plot_ppc_dtypes(get_result_list):
    for name, result in zip(*get_result_list):
        for y_type in ["counts", "countrate", "photon_flux", "photon_flux_density"]:
            result.plot_ppc(y_type=y_type)


def test_plot_corner(get_result_list):
    for name, result in zip(*get_result_list):
        result.plot_corner()


def test_table(request, get_result_list):
    print(request.node.name)
    for name, result in zip(*get_result_list):
        print(result.table())


def test_compare(request, get_result_list):
    plot_corner_comparison({name: res for name, res in zip(*get_result_list)})
    plt.suptitle(request.node.name)
    plt.show()


def test_posterior_photon_flux(get_joint_mcmc_result):
    result = get_joint_mcmc_result[0]
    e_min, e_max = 0.7, 1.2
    result.photon_flux(e_min, e_max, register=True)
    assert f"photon_flux_{e_min:.1f}_{e_max:.1f}" in list(result.inference_data.posterior.keys())


def test_posterior_energy_flux(get_joint_mcmc_result):
    result = get_joint_mcmc_result[0]
    e_min, e_max = 0.7, 1.2
    result.energy_flux(e_min, e_max, register=True)
    assert f"energy_flux_{e_min:.1f}_{e_max:.1f}" in list(result.inference_data.posterior.keys())


def test_posterior_luminosity(get_joint_mcmc_result):
    result = get_joint_mcmc_result[0]
    e_min, e_max = 0.7, 1.2

    with pytest.raises(ValueError):
        result.luminosity(e_min, e_max, register=True)

    with pytest.raises(ValueError):
        result.luminosity(e_min, e_max, distance=10 * u.kpc, redshift=0.1, register=True)

    result.luminosity(e_min, e_max, redshift=0.1, register=True)

    assert f"luminosity_{e_min:.1f}_{e_max:.1f}" in list(result.inference_data.posterior.keys())



import numpy as np
import xarray as xr

from .ogip import DataPHA


class Observation(xr.Dataset):
    """
    Class to store the data of an observation
    """

    counts: xr.DataArray
    """The observed counts"""
    folded_counts: xr.DataArray
    """The observed counts, after grouping"""
    grouping: xr.DataArray
    """The grouping matrix"""
    quality: xr.DataArray
    """The quality flag"""
    exposure: xr.DataArray
    """The total exposure"""
    background: xr.DataArray
    """The background counts if provided, otherwise 0"""
    folded_background: xr.DataArray
    """The background counts, after grouping"""

    __slots__ = (
        "grouping",
        "channel",
        "quality",
        "exposure",
        "background",
        "folded_background",
        "counts",
        "folded_counts",
    )

    _default_attributes = {"description": "X-ray observation dataset"}

    @classmethod
    def from_matrix(
        cls,
        counts,
        grouping,
        channel,
        quality,
        exposure,
        background=None,
        background_unscaled=None,
        backratio=1.0,
        attributes: dict | None = None,
    ):
        if attributes is None:
            attributes = {}

        if background is None or background_unscaled is None:
            background = np.zeros_like(counts, dtype=np.int64)
            background_unscaled = np.zeros_like(counts, dtype=np.int64)

        data_dict = {
            "counts": (
                ["instrument_channel"],
                np.asarray(counts, dtype=np.int64),
                {"description": "Counts", "unit": "photons"},
            ),
            "folded_counts": (
                ["folded_channel"],
                np.asarray(np.ma.filled(grouping @ counts), dtype=np.int64),
                {"description": "Folded counts, after grouping", "unit": "photons"},
            ),
            "grouping": (
                ["folded_channel", "instrument_channel"],
                grouping,
                {"description": "Grouping matrix."},
            ),
            "quality": (
                ["instrument_channel"],
                np.asarray(quality, dtype=np.int64),
                {"description": "Quality flag."},
            ),
            "exposure": ([], float(exposure), {"description": "Total exposure", "unit": "s"}),
            "backratio": (
                ["instrument_channel"],
                np.asarray(backratio, dtype=float),
                {"description": "Background scaling (SRC_BACKSCAL/BKG_BACKSCAL)"},
            ),
            "folded_backratio": (
                ["folded_channel"],
                np.asarray(np.ma.filled(grouping @ backratio), dtype=float),
                {"description": "Background scaling after grouping"},
            ),
            "background": (
                ["instrument_channel"],
                np.asarray(background, dtype=np.int64),
                {"description": "Background counts", "unit": "photons"},
            ),
            "folded_background_unscaled": (
                ["folded_channel"],
                np.asarray(np.ma.filled(grouping @ background_unscaled), dtype=np.int64),
                {"description": "Background counts", "unit": "photons"},
            ),
            "folded_background": (
                ["folded_channel"],
                np.asarray(np.ma.filled(grouping @ background), dtype=np.float64),
                {"description": "Background counts", "unit": "photons"},
            ),
        }

        return cls(
            data_dict,
            coords={
                "channel": (
                    ["instrument_channel"],
                    np.asarray(channel, dtype=np.int64),
                    {"description": "Channel number"},
                ),
                "grouped_channel": (
                    ["folded_channel"],
                    np.arange(len(grouping @ counts), dtype=np.int64),
                    {"description": "Channel number"},
                ),
            },
            attrs=cls._default_attributes
            if attributes is None
            else attributes | cls._default_attributes,
        )

    @classmethod
    def from_ogip_container(cls, pha: DataPHA, bkg: DataPHA | None = None, **metadata):
        if bkg is not None:
            backratio = np.nan_to_num(
                (pha.backscal * pha.exposure * pha.areascal)
                / (bkg.backscal * bkg.exposure * bkg.areascal)
            )
        else:
            backratio = np.ones_like(pha.counts)

        if (bkg is not None) and ("NET" in pha.flags):
            counts = pha.counts + bkg.counts * backratio
        else:
            counts = pha.counts

        return cls.from_matrix(
            counts,
            pha.grouping,
            pha.channel,
            pha.quality,
            pha.exposure,
            backratio=backratio,
            background=bkg.counts * backratio if bkg is not None else None,
            background_unscaled=bkg.counts if bkg is not None else None,
            attributes=metadata,
        )

    @classmethod
    def from_pha_file(cls, pha_path: str, bkg_path: str | None = None, **metadata):
        """
        Build an observation from a PHA file

        Parameters:
            pha_path : Path to the PHA file
            bkg_path : Path to the background file
            metadata : Additional metadata to add to the observation
        """
        from .util import data_path_finder

        arf_path, rmf_path, bkg_path_default = data_path_finder(
            pha_path, require_arf=False, require_rmf=False, require_bkg=False
        )
        bkg_path = bkg_path_default if bkg_path is None else bkg_path

        pha = DataPHA.from_file(pha_path)
        bkg = DataPHA.from_file(bkg_path) if bkg_path is not None else None

        if metadata is None:
            metadata = {}

        metadata.update(
            observation_file=pha_path,
            background_file=bkg_path,
            response_matrix_file=rmf_path,
            ancillary_response_file=arf_path,
        )

        return cls.from_ogip_container(pha, bkg=bkg, **metadata)

    def plot_counts(self, **kwargs):
        """
        Plot the counts

        Parameters:
            **kwargs : `kwargs` passed to https://docs.xarray.dev/en/latest/generated/xarray.DataArray.plot.step.html#xarray.DataArray.plot.line
        """

        return self.counts.plot.step(x="instrument_channel", yscale="log", where="post", **kwargs)

    def plot_grouping(self):
        """
        Plot the grouping matrix and compare the grouped counts to the true counts
        in the original channels.
        """

        import matplotlib.pyplot as plt
        import seaborn as sns

        fig = plt.figure(figsize=(6, 6))
        gs = fig.add_gridspec(
            2,
            2,
            width_ratios=(4, 1),
            height_ratios=(1, 4),
            left=0.1,
            right=0.9,
            bottom=0.1,
            top=0.9,
            wspace=0.05,
            hspace=0.05,
        )
        ax = fig.add_subplot(gs[1, 0])
        ax_histx = fig.add_subplot(gs[0, 0], sharex=ax)
        ax_histy = fig.add_subplot(gs[1, 1], sharey=ax)
        sns.heatmap(self.grouping.data.todense().T, ax=ax, cbar=False)
        ax_histx.step(np.arange(len(self.folded_counts)), self.folded_counts, where="post")
        ax_histy.step(self.counts, np.arange(len(self.counts)), where="post")

        ax.set_xlabel("Grouped channels")
        ax.set_ylabel("Channels")
        ax_histx.set_ylabel("Grouped counts")
        ax_histy.set_xlabel("Counts")

        ax_histx.semilogy()
        ax_histy.semilogx()

        _ = [label.set_visible(False) for label in ax_histx.get_xticklabels()]
        _ = [label.set_visible(False) for label in ax_histy.get_yticklabels()]



import inspect

from .abc import ModelComponent
from .additive import *  # noqa: F403
from .multiplicative import *  # noqa: F403


def all_models(cls: ModelComponent) -> list[ModelComponent]:
    """
    Return a list of all the subclasses of a given ModelComponent class
    """
    subclasses = list(
        set(cls.__subclasses__()).union([s for c in cls.__subclasses__() for s in all_models(c)])
    )

    return [s for s in subclasses if not inspect.isabstract(s)]


# I want to put all the existent models in the namespace to build them in the good ol' XSPEC way
model_components = {cls.__name__: cls for cls in all_models(ModelComponent)}
additive_components = {
    cls.__name__: cls
    for cls in all_models(AdditiveComponent)  # noqa: F405
}
multiplicative_components = {
    cls.__name__: cls
    for cls in all_models(MultiplicativeComponent)  # noqa: F405
}



import os

import numpy as np
import pytest

from jaxspec.util.online_storage import table_manager

xspec = pytest.importorskip("xspec")


@pytest.fixture
def load_jaxspec_data(request, monkeypatch):
    from jaxspec.data import ObsConfiguration, Observation

    file_pha = table_manager.fetch("example_data/NGC7793_ULX4/PN_spectrum_grp20.fits")
    pn_rmf_path = table_manager.fetch("example_data/NGC7793_ULX4/PN.rmf")
    pn_arf_path = table_manager.fetch("example_data/NGC7793_ULX4/PN.arf")
    dir_path = os.path.dirname(file_pha)
    monkeypatch.chdir(os.path.join(os.path.dirname(request.fspath.dirname), dir_path))

    low_energy, high_energy = 0.5, 8.0
    folding = ObsConfiguration.from_pha_file(
        file_pha, low_energy=low_energy, high_energy=high_energy
    )
    observation = Observation.from_pha_file(file_pha)

    return folding, observation


@pytest.fixture
def load_xspec_data(request, monkeypatch):
    low_energy, high_energy = 0.5, 8.0

    file_pha = table_manager.fetch("example_data/NGC7793_ULX4/PN_spectrum_grp20.fits")
    pn_rmf_path = table_manager.fetch("example_data/NGC7793_ULX4/PN.rmf")
    pn_arf_path = table_manager.fetch("example_data/NGC7793_ULX4/PN.arf")
    dir_path = os.path.dirname(file_pha)
    monkeypatch.chdir(os.path.join(os.path.dirname(request.fspath.dirname), dir_path))

    xspec.AllData.clear()
    xspec.Plot.xAxis = "keV"
    xspec_observation = xspec.Spectrum(file_pha)
    xspec_observation.ignore(f"0.0-{low_energy:.1f} {high_energy:.1f}-**")

    return xspec_observation


def test_obs_constitantcy(load_xspec_data, load_jaxspec_data):
    file_pha = table_manager.fetch("example_data/NGC7793_ULX4/PN_spectrum_grp20.fits")
    xspec_observation = load_xspec_data
    folding, observation = load_jaxspec_data

    assert np.isclose(
        xspec_observation.exposure, float(observation.exposure), 0.1
    ), f"Exposure is the same as loaded by XSPEC for {file_pha}"
    assert len(xspec_observation.values) == len(
        folding.coords["folded_channel"]
    ), f"The number of grouped channel is not the same as XSPEC {file_pha}"


def test_bins(load_xspec_data, load_jaxspec_data):
    file_pha = table_manager.fetch("example_data/NGC7793_ULX4/PN_spectrum_grp20.fits")
    xspec_observation = load_xspec_data
    folding, observation = load_jaxspec_data

    xspec_in_energies = np.vstack(
        [
            np.asarray(xspec_observation.response.energies)[0:-1],
            np.asarray(xspec_observation.response.energies)[1:],
        ]
    )

    xspec_out_energies = np.asarray(xspec_observation.energies).T

    assert np.isclose(
        folding.out_energies, xspec_out_energies
    ).all(), f"The grouped channel energy bins are not the same as XSPEC {file_pha}"
    assert np.isclose(
        folding.in_energies, xspec_in_energies
    ).all(), f"The unfolded channel energy bins are not the same as XSPEC {file_pha}"



# ruff: noqa: E402

import chex

chex.set_n_cpu_devices(n=4)

import os

from pathlib import Path

import jax.numpy as jnp
import numpyro
import numpyro.distributions as dist
import pooch
import pytest
import yaml

from jax import config
from jaxspec.data.util import load_example_obsconf
from jaxspec.fit import MCMCFitter
from jaxspec.model.additive import Blackbodyrad, Powerlaw
from jaxspec.model.multiplicative import Tbabs

config.update("jax_enable_x64", True)
numpyro.set_platform("cpu")
numpyro.set_host_device_count(4)

prior_shared_pars = {
    "powerlaw_1_alpha": dist.Uniform(0, 5),
    "powerlaw_1_norm": dist.LogUniform(1e-5, 1e-2),
    "blackbodyrad_1_kT": dist.Uniform(0, 5),
    "blackbodyrad_1_norm": dist.LogUniform(1e-2, 1e2),
    "tbabs_1_nh": dist.Uniform(0, 1),
}

prior_split_pars = {
    "powerlaw_1_alpha": dist.Uniform(0, 5),
    "powerlaw_1_norm": dist.LogUniform(1e-5 * jnp.ones(3), 1e-2 * jnp.ones(3)),
    "blackbodyrad_1_kT": dist.Uniform(0, 5),
    "blackbodyrad_1_norm": dist.LogUniform(1e-2, 1e2),
    "tbabs_1_nh": dist.Uniform(0, 1),
}

prior = {
    "powerlaw_1_alpha": dist.Uniform(0, 5),
    "powerlaw_1_norm": dist.LogUniform(1e-5, 1e-2),
    "blackbodyrad_1_kT": dist.Uniform(0, 5),
    "blackbodyrad_1_norm": dist.LogUniform(1e-2, 1e2),
    "tbabs_1_nh": dist.Uniform(0, 1),
}

single_obsconf = load_example_obsconf("NGC7793_ULX4_PN")
list_of_obsconf = list(load_example_obsconf("NGC7793_ULX4_ALL").values())
dict_of_obsconf = load_example_obsconf("NGC7793_ULX4_ALL")

# Dir containing 8 files
parent_directory = Path(__file__).parent.resolve()
data_directory = parent_directory / "data"

if not data_directory.exists():
    os.mkdir(data_directory)

with open(parent_directory / "data_files.yml") as file:
    data_collection = yaml.safe_load(file)

with open(parent_directory / "data_hash.yml") as file:
    data_hash = yaml.safe_load(file)

pooch_dataset = pooch.create(
    base_url="https://github.com/HEACIT/curated-test-data/raw/main/",
    path=str(data_directory),
    registry=data_hash,
)

for file in data_hash.keys():
    pooch_dataset.fetch(file)


@pytest.fixture(scope="session")
def obsconfs():
    return list(load_example_obsconf("NGC7793_ULX4_ALL").values())


@pytest.fixture(scope="session")
def observations():
    from jaxspec.data.util import load_example_pha

    return list(load_example_pha("NGC7793_ULX4_ALL").values())


@pytest.fixture(scope="session")
def instruments():
    from jaxspec.data.util import load_example_instruments

    return list(load_example_instruments("NGC7793_ULX4_ALL").values())


@pytest.fixture(scope="session")
def obs_model_prior(obsconfs):
    from jaxspec.model.additive import Blackbodyrad, Powerlaw
    from jaxspec.model.multiplicative import Tbabs

    model = Tbabs() * (Powerlaw() + Blackbodyrad())

    return obsconfs, model, prior


@pytest.fixture(scope="session")
def get_individual_mcmc_results(obs_model_prior):
    obsconfs, model, prior = obs_model_prior

    return [MCMCFitter(model, prior, obsconf).fit(num_samples=5000) for obsconf in obsconfs]


@pytest.fixture(scope="session")
def get_joint_mcmc_result(obs_model_prior):
    obsconfs, model, prior = obs_model_prior

    return [MCMCFitter(model, prior, obsconfs).fit(num_samples=5000)]


@pytest.fixture(scope="session")
def get_result_list(get_individual_mcmc_results, get_joint_mcmc_result):
    result_list = []
    result_list += get_individual_mcmc_results
    result_list += get_joint_mcmc_result

    name_list = []
    name_list += ["PN_mcmc", "MOS1_mcmc", "MOS2_mcmc"]
    name_list += ["Joint_mcmc"]

    return name_list, result_list


spectral_model = Tbabs() * (Powerlaw() + Blackbodyrad())


@pytest.fixture(scope="session")
def prior_with_free_pars_on_obs():
    """
    Prior distribution with a free normalisation for each observation
    """

    return


@pytest.fixture(scope="session")
@pytest.mark.parametrize("obsconf, model, prior", ["NGC7793_ULX4_ALL"])
def fitting_setup_multiple_obs_and_free_pars(obsconf, model, prior):
    model = Tbabs() * (Powerlaw() + Blackbodyrad())

    return obsconfs, model, prior



from typing import Dict
from .results import FitResult
from chainconsumer import ChainConsumer


def plot_corner_comparison(obs_dict: Dict[str, FitResult], **kwargs):
    """
    Plot the correlation plot of parameters from different fitted observations. Observations are passed in as a
    dictionary. Each observation is named according to its key. It shall be used to compare the same model independently
    fitted to different observations.

    Parameters:
        obs_dict: a dictionary containing the observations to plot. The keys are the names of the observations.
    """

    c = ChainConsumer()

    for name, obs in obs_dict.items():
        c.add_chain(obs.to_chain(name))

    return c.plotter.plot(**kwargs)



from __future__ import annotations

import catppuccin
import matplotlib.pyplot as plt
import numpy as np

from astropy import units as u
from catppuccin.extras.matplotlib import load_color
from cycler import cycler
from jax.typing import ArrayLike
from scipy.integrate import trapezoid
from scipy.stats import nbinom, norm

from jaxspec.data import ObsConfiguration

PALETTE = catppuccin.PALETTE.latte

COLOR_CYCLE = [
    load_color(PALETTE.identifier, color)
    for color in ["sky", "teal", "green", "yellow", "peach", "maroon", "red", "pink", "mauve"][::-1]
]

LINESTYLE_CYCLE = ["dashed", "dotted", "dashdot", "solid"]

SPECS_CYCLE = cycler(linestyle=LINESTYLE_CYCLE) * cycler(color=COLOR_CYCLE)

SPECTRUM_COLOR = load_color(PALETTE.identifier, "blue")
SPECTRUM_DATA_COLOR = load_color(PALETTE.identifier, "overlay2")
BACKGROUND_COLOR = load_color(PALETTE.identifier, "sapphire")
BACKGROUND_DATA_COLOR = load_color(PALETTE.identifier, "overlay0")


def sigma_to_percentile_intervals(sigmas):
    intervals = []
    for sigma in sigmas:
        lower_bound = 100 * norm.cdf(-sigma)
        upper_bound = 100 * norm.cdf(sigma)
        intervals.append((lower_bound, upper_bound))
    return intervals


def _plot_poisson_data_with_error(
    ax: plt.Axes,
    x_bins: ArrayLike,
    y: ArrayLike,
    y_low: ArrayLike,
    y_high: ArrayLike,
    color=SPECTRUM_DATA_COLOR,
    linestyle="none",
    alpha=0.3,
):
    """
    Plot Poisson data with error bars. We extrapolate the intrinsic error of the observation assuming a prior rate
    distributed according to a Gamma RV.
    """

    ax_to_plot = ax.errorbar(
        np.sqrt(x_bins[0] * x_bins[1]),
        y,
        xerr=np.abs(x_bins - np.sqrt(x_bins[0] * x_bins[1])),
        yerr=[
            np.maximum(y - y_low, 0),
            np.maximum(y_high - y, 0),
        ],
        color=color,
        linestyle=linestyle,
        alpha=alpha,
        capsize=2,
    )

    return ax_to_plot


def _plot_binned_samples_with_error(
    ax: plt.Axes,
    x_bins: ArrayLike,
    y_samples: ArrayLike,
    color=SPECTRUM_COLOR,
    alpha_median: float = 0.7,
    alpha_envelope: (float, float) = (0.15, 0.25),
    linestyle="solid",
    n_sigmas=3,
):
    """
    Helper function to plot the posterior predictive distribution of the model. The function
    computes the percentiles of the posterior predictive distribution and plot them as a shaded
    area. If the observed data is provided, it is also plotted as a step function.

    Parameters:
        x_bins: The bin edges of the data (2 x N).
        y_samples: The samples of the posterior predictive distribution (Samples X N).
        ax: The matplotlib axes object.
        color: The color of the posterior predictive distribution.
    """

    median = ax.stairs(
        list(np.median(y_samples, axis=0)),
        edges=[*list(x_bins[0]), x_bins[1][-1]],
        color=color,
        alpha=alpha_median,
        linestyle=linestyle,
    )

    # The legend cannot handle fill_between, so we pass a fill to get a fancy icon
    (envelope,) = ax.fill(np.nan, np.nan, alpha=alpha_envelope[-1], facecolor=color)

    if n_sigmas == 1:
        alpha_envelope = (alpha_envelope[1], alpha_envelope[0])

    for percentile, alpha in zip(
        sigma_to_percentile_intervals(list(range(n_sigmas, 0, -1))),
        np.linspace(*alpha_envelope, n_sigmas),
    ):
        percentiles = np.percentile(y_samples, percentile, axis=0)
        ax.stairs(
            percentiles[1],
            edges=[*list(x_bins[0]), x_bins[1][-1]],
            baseline=percentiles[0],
            alpha=alpha,
            fill=True,
            color=color,
        )

    return [(median, envelope)]


def _compute_effective_area(
    obsconf: ObsConfiguration,
    x_unit: str | u.Unit = "keV",
):
    """
    Helper function to compute the bins and effective area of an observational configuration

    Parameters:
        obsconf: The observational configuration.
        x_unit: The unit of the x-axis. It can be either a string (parsable by astropy.units) or an astropy unit. It must be homogeneous to either a length, a frequency or an energy.
    """

    # Note to Simon : do not change xbins[1] - xbins[0] to
    # np.diff, you already did this twice and forgot that it does not work since diff keeps the dimensions
    # and enable weird broadcasting that makes the plot fail

    xbins = obsconf.out_energies * u.keV
    xbins = xbins.to(x_unit, u.spectral())

    # This computes the total effective area within all bins
    # This is a bit weird since the following computation is equivalent to ignoring the RMF
    exposure = obsconf.exposure.data * u.s
    mid_bins_arf = obsconf.in_energies.mean(axis=0) * u.keV
    mid_bins_arf = mid_bins_arf.to(x_unit, u.spectral())
    e_grid = np.linspace(*xbins, 10)
    interpolated_arf = np.interp(e_grid.value, mid_bins_arf.value, obsconf.area)
    integrated_arf = (
        trapezoid(interpolated_arf, x=e_grid.value, axis=0)
        / (
            np.abs(
                xbins[1] - xbins[0]
            ).value  # Must fold in abs because some units reverse the ordering of the bins
        )
        * u.cm**2
    )

    return xbins, exposure, integrated_arf


def _error_bars_for_observed_data(observed_counts, denominator, units, sigma=1):
    r"""
    Compute the error bars for the observed data assuming a prior Gamma distribution

    Parameters:
        observed_counts: array of integer counts
        denominator: normalization factor (e.g. effective area)
        units: unit to convert to
        sigma: dispersion to use for quantiles computation

    Returns:
        y_observed: observed counts in the desired units
        y_observed_low: lower bound of the error bars
        y_observed_high: upper bound of the error bars
    """

    percentile = sigma_to_percentile_intervals([sigma])[0]

    y_observed = (observed_counts * u.ct / denominator).to(units)

    y_observed_low = (
        nbinom.ppf(percentile[0] / 100, observed_counts, 0.5) * u.ct / denominator
    ).to(units)

    y_observed_high = (
        nbinom.ppf(percentile[1] / 100, observed_counts, 0.5) * u.ct / denominator
    ).to(units)

    return y_observed, y_observed_low, y_observed_high



import numpy as np

from scipy.interpolate import RegularGridInterpolator


class RegularGridInterpolatorWithGrad(RegularGridInterpolator):
    """
    A subclass of SciPy's RegularGridInterpolator that also returns the gradient
    of each interpolated output with respect to input coordinates.

    Supports:
        - Linear interpolation
        - Out-of-bounds handling (fill_value=0)
        - Multi-dimensional output (e.g., RGB or vector fields)
        - Batched or single-point evaluation

    Returns:
        - values: shape (..., output_dim)
        - gradients: shape (..., input_dim, output_dim)
    """

    def __init__(self, points, values, **kwargs):
        kwargs.setdefault("method", "linear")
        kwargs.setdefault("bounds_error", False)
        kwargs.setdefault("fill_value", 0.0)
        self.points = [np.asarray(p) for p in points]
        self.input_dim = len(self.points)

        self.output_shape = values.shape[self.input_dim :]
        values_reshaped = values.reshape(*[len(p) for p in self.points], -1)  # flatten output

        super().__init__(self.points, values_reshaped, **kwargs)

    def __call__(self, xi, return_gradient=True):
        xi = np.atleast_2d(xi).astype(float)
        n_points, n_dims = xi.shape
        assert n_dims == self.input_dim, "Dim mismatch"

        # Interpolate values
        flat_vals = super().__call__(xi)  # shape: (n_points, output_dim)
        values = flat_vals.reshape(n_points, *self.output_shape)

        if not return_gradient:
            return values[0] if values.shape[0] == 1 else values

        gradients = np.zeros((n_points, self.input_dim, np.prod(self.output_shape)), dtype=float)

        for d, grid in enumerate(self.points):
            xq = xi[:, d]
            idx_upper = np.searchsorted(grid, xq, side="right")
            idx_lower = idx_upper - 1

            idx_lower = np.clip(idx_lower, 0, len(grid) - 2)
            idx_upper = np.clip(idx_upper, 1, len(grid) - 1)

            xi_low = xi.copy()
            xi_high = xi.copy()
            xi_low[:, d] = grid[idx_lower]
            xi_high[:, d] = grid[idx_upper]

            f_low = super().__call__(xi_low)
            f_high = super().__call__(xi_high)
            delta = (grid[idx_upper] - grid[idx_lower])[:, np.newaxis]

            grad = np.where(delta != 0, (f_high - f_low) / delta, 0.0)
            gradients[:, d, :] = grad

        # Reshape output properly
        gradients = gradients.reshape(n_points, self.input_dim, *self.output_shape)

        if values.shape[0] == 1:
            return values[0], gradients[0]
        else:
            return values, gradients



from collections.abc import Callable
from typing import TYPE_CHECKING

import jax
import jax.numpy as jnp
import numpy as np
import numpyro

from jax.experimental.sparse import BCOO
from jax.typing import ArrayLike
from numpyro.distributions import Distribution

if TYPE_CHECKING:
    from jaxspec.data import ObsConfiguration
    from jaxspec.model.abc import SpectralModel
    from jaxspec.util.typing import PriorDictType


class TiedParameter:
    def __init__(self, tied_to, func):
        self.tied_to = tied_to
        self.func = func


def forward_model(
    model: "SpectralModel",
    parameters,
    obs_configuration: "ObsConfiguration",
    sparse=False,
    gain: Callable | None = None,
    shift: Callable | None = None,
    split_branches: bool = False,
    n_points: int | None = 2,
):
    energies = np.asarray(obs_configuration.in_energies)

    if sparse:
        # folding.transfer_matrix.data.density > 0.015 is a good criterion to consider sparsify
        transfer_matrix = BCOO.from_scipy_sparse(
            obs_configuration.transfer_matrix.data.to_scipy_sparse().tocsr()
        )

    else:
        transfer_matrix = np.asarray(obs_configuration.transfer_matrix.data.todense())

    energies = shift(energies) if shift is not None else energies
    energies = jnp.clip(energies, min=1e-6)  # Ensure shifted energies remain positive
    factor = gain(energies) if gain is not None else 1.0
    factor = jnp.clip(factor, min=0.0)  # Ensure the gain is positive to avoid NaNs

    if not split_branches:
        expected_counts = transfer_matrix @ (
            model.photon_flux(parameters, *energies, n_points=n_points) * factor
        )
        return jnp.clip(expected_counts, min=1e-6)  # Ensure the expected counts are positive

    else:
        model_flux = model.photon_flux(
            parameters, *energies, split_branches=True, n_points=n_points
        )
        return jax.tree.map(
            lambda f: jnp.clip(transfer_matrix @ (f * factor), min=1e-6), model_flux
        )


def build_prior(prior: "PriorDictType", expand_shape: tuple = (), prefix=""):
    """
    Transform a dictionary of prior distributions into a dictionary of parameters sampled from the prior.
    Must be used within a numpyro model.
    """
    parameters = {}
    params_to_tie = {}

    for key, value in prior.items():
        # Split the key to extract the module name and parameter name
        module_name, param_name = key.rsplit("_", 1)

        if isinstance(value, Distribution):
            parameters[key] = jnp.ones(expand_shape) * numpyro.sample(
                f"{prefix}{module_name}_{param_name}", value
            )

        elif isinstance(value, TiedParameter):
            params_to_tie[key] = value

        elif isinstance(value, ArrayLike):
            parameters[key] = jnp.ones(expand_shape) * value

        else:
            raise ValueError(
                f"Invalid prior type {type(value)} for parameter {prefix}{module_name}_{param_name} : {value}"
            )

    for key, value in params_to_tie.items():
        func_to_apply = value.func
        tied_to = value.tied_to
        parameters[key] = func_to_apply(parameters[tied_to])

    return parameters



import jax
import jax.numpy as jnp
import numpy as np
import xarray as xr

from astropy.table import Table
from flax import nnx
from jax import lax
from jax.scipy.interpolate import RegularGridInterpolator
from jax.typing import ArrayLike
from tqdm.auto import tqdm

from ..model.abc import AdditiveComponent
from .interpolator import RegularGridInterpolatorWithGrad


class TableManager:
    """
    Handler for the tabulated data from `xspec` tabulated models. The table must follow the format speficied in the
    [`atable`](https://heasarc.gsfc.nasa.gov/docs/xanadu/xspec/manual/XSmodelAtable.html) additive component.
    """

    def __init__(self, table_path):
        self._table_path = table_path

        # Read the tables once.
        raw_parameters = Table.read(self._table_path, "PARAMETERS")
        raw_energies = Table.read(self._table_path, "ENERGIES")
        raw_spectra = Table.read(self._table_path, "SPECTRA")

        # Build and wrap the parameter table.
        parameter_table = tuple(np.asarray(row["VALUE"], dtype=float) for row in raw_parameters)
        parameters = []
        for row in raw_parameters:
            parameter_name = row["NAME"].rstrip()
            parameters.append((parameter_name, float(row["INITIAL"])))

        # Process and wrap the energies table.
        energy_low = np.asarray(raw_energies["ENERG_LO"].value, dtype=float)
        energy_high = np.asarray(raw_energies["ENERG_HI"].value, dtype=float)
        energies_table = np.vstack((energy_low, energy_high))

        # Compute the parameter shape
        parameter_shape = tuple()
        for parameter in parameter_table:
            parameter_shape += parameter.shape

        # Build and wrap the spectra table.
        total_shape = parameter_shape + energy_low.shape

        spectra_table = np.empty(total_shape, dtype=float)
        for idx, par_indexes in enumerate(np.ndindex(*parameter_shape)):
            spectra_table[par_indexes] = np.asarray(raw_spectra["INTPSPEC"][idx], dtype=float)

        self.parameters = parameters
        self.energies_table = energies_table
        self.parameter_table = parameter_table
        self.parameter_shape = parameter_shape
        self.spectra_table = spectra_table

    def check_proper_ordering(self):
        """
        Assert that the parameter ordering in the spectra table is consistent with the parameter table.
        """

        spectra_table = Table.read(self._table_path, "SPECTRA")
        pars_mesh = np.meshgrid(*self.parameter_table, indexing="ij", sparse=False)

        for idx, par_indexes in enumerate(
            tqdm(np.ndindex(*self.parameter_shape), total=len(spectra_table))
        ):
            expected = np.asarray([par[par_indexes] for par in pars_mesh])
            obtained = jnp.asarray(spectra_table["PARAMVAL"][idx], dtype=float)
            assert jnp.allclose(expected, obtained), "Parameter ordering mismatch"


class TabulatedModel(nnx.Module):
    """
    See https://heasarc.gsfc.nasa.gov/docs/heasarc/caldb/docs/memos/ogip_92_009/ogip_92_009.pdf
    """

    def __init__(self, table_path: str):
        table_manager = TableManager(table_path)

        self._parameters = table_manager.parameters
        self._parameter_table = table_manager.parameter_table
        self._spectra_table = table_manager.spectra_table
        self._energies_table = table_manager.energies_table

        # Instantiate the parameters of the nnx.Module
        for parameter_name, value in table_manager.parameters:
            setattr(self, parameter_name, nnx.Param(value))

        del table_manager

    def get_parameter_list(self):
        parameter_list = []
        for parameter_name, _ in self._parameters:
            parameter_list.append(getattr(self, parameter_name))

        return jnp.asarray(parameter_list, dtype=float)


"""
import numpy as np
from scipy.sparse import csr_matrix

def redistribution_matrix(old_e_low, old_e_high, new_e_low, new_e_high):
    new_bins = jnp.stack([new_e_low, new_e_high], axis=1)

    def scan_body(carry, new_bin):
        new_low, new_high = new_bin

        # Compute the overlap between the current new bin and all old bins.
        lower_bounds = jnp.maximum(old_e_low, new_low)
        upper_bounds = jnp.minimum(old_e_high, new_high)
        overlap_fraction = jnp.maximum(0, upper_bounds - lower_bounds) / (old_e_high - old_e_low)

        return carry, overlap_fraction

    _, matrix = lax.scan(scan_body, None, new_bins)
    return matrix.T
"""


def redistribute(
    integrated_spectrum: ArrayLike,
    old_e_low: ArrayLike,
    old_e_high: ArrayLike,
    e_low: ArrayLike,
    e_high: ArrayLike,
) -> ArrayLike:
    """
    Redistribute the integrated spectrum over the new energy bins.

    Parameters:
        integrated_spectrum: Integrated spectrum to redistribute.
        old_e_low: Lower bounds of the old energy bins.
        old_e_high: Upper bounds of the old energy bins.
        e_low: Lower bounds of the new energy bins.
        e_high: Upper bounds of the new energy bins.
    """
    new_bins = jnp.stack([e_low, e_high], axis=1)

    def scan_body(carry, new_bin):
        new_low, new_high = new_bin

        # Compute the overlap between the current new bin and all old bins.
        lower_bounds = jnp.maximum(old_e_low, new_low)
        upper_bounds = jnp.minimum(old_e_high, new_high)
        overlap_fraction = jnp.maximum(0, upper_bounds - lower_bounds) / (old_e_high - old_e_low)

        # Sum over old bins: each old bin contributes its integrated value times
        # the fraction of the new bin that overlaps with it.
        new_intensity = jnp.sum(overlap_fraction * integrated_spectrum)

        return carry, new_intensity

    _, redistributed_values = lax.scan(scan_body, None, new_bins)
    return redistributed_values


class AdditiveTabulated(AdditiveComponent, TabulatedModel):
    """
    Equivalent of the [`atable`](https://heasarc.gsfc.nasa.gov/docs/xanadu/xspec/manual/XSmodelAtable.html) additive
    component in `xspec`.
    """

    def __init__(self, table_path: str):
        super().__init__(table_path)

        self.norm = nnx.Param(1.0)

        self._interpolator = RegularGridInterpolator(
            self._parameter_table, self._spectra_table, fill_value=0.0
        )

    def _integrate_on_grid(self):
        return self._interpolator(self.get_parameter_list()).squeeze()

    def integrated_continuum(self, e_low, e_high):
        integrated_spectrum = self._integrate_on_grid()
        return jnp.asarray(self.norm) * redistribute(
            integrated_spectrum, *self._energies_table, e_low, e_high
        )


class TabulatedModelXarray(nnx.Module):
    """
    See https://heasarc.gsfc.nasa.gov/docs/heasarc/caldb/docs/memos/ogip_92_009/ogip_92_009.pdf
    """

    def __init__(self, table_path: str):
        ds = xr.open_dataset(table_path)

        self._parameters = [dim for dim in ds["spectra"].dims if dim != "energy"]
        self._parameter_table = [np.asarray(ds[key]) for key in self._parameters]
        self._spectra_table = ds["spectra"]
        self._energies_table = (np.asarray(ds["energy_low"]), np.asarray(ds["energy_high"]))

        # Instantiate the parameters of the nnx.Module
        for parameter_name, parameter_value in zip(self._parameters, self._parameter_table):
            setattr(
                self,
                parameter_name,
                nnx.Param(np.median(parameter_value) + np.random.uniform(-5e-1, +5e-1)),
            )

    def get_parameter_list(self):
        parameter_list = []
        for parameter_name in self._parameters:
            parameter_list.append(getattr(self, parameter_name))

        return jnp.asarray(parameter_list, dtype=float)


class AdditiveTabulatedXarray(AdditiveComponent, TabulatedModelXarray):
    def __init__(self, table_path: str):
        super().__init__(table_path)

        self.norm = nnx.Param(1.0)

        interpolator = RegularGridInterpolatorWithGrad(
            self._parameter_table, np.asarray(self._spectra_table)
        )

        def callback(pars):
            value, grad = interpolator(pars)
            return np.vstack([value[None, :], grad])

        result = callback(self.get_parameter_list())

        out_type = jax.ShapeDtypeStruct(jnp.shape(result), jnp.result_type(result))

        @jax.custom_jvp
        def spectrum_interpolation(pars):
            result = jax.pure_callback(
                lambda p: callback(p), out_type, pars, vmap_method="legacy_vectorized"
            )
            return result[0, ...]

        @spectrum_interpolation.defjvp
        def spectrum_interpolation_jvp(primals, tangents):
            pars = primals
            pars_dot = tangents

            result = jax.pure_callback(
                lambda p: callback(p), out_type, pars, vmap_method="legacy_vectorized"
            )
            value = result[0, ...]
            grad = result[1:, ...]

            return value, jnp.squeeze(jnp.asarray(pars_dot) @ grad)

        self._spectrum_interpolation = spectrum_interpolation

    def _integrate_on_grid(self):
        return self._spectrum_interpolation(jnp.asarray(self.get_parameter_list()))

    def integrated_continuum(self, e_low, e_high):
        integrated_spectrum = self._integrate_on_grid()
        return jnp.asarray(self.norm) * redistribute(
            integrated_spectrum, *self._energies_table, e_low, e_high
        )



import os
import sys

import chex
import jax
import pytest

from jaxspec.data import ObsConfiguration
from jaxspec.data.util import fakeit_for_multiple_parameters

chex.set_n_cpu_devices(n=4)

current_dir = os.path.dirname(os.path.abspath(__file__))
source_dir = os.path.abspath(os.path.join(current_dir, ".."))
sys.path.append(source_dir)


@pytest.fixture
def multidimensional_parameters():
    from numpy.random import default_rng

    rng = default_rng(42)

    num_params = (8, 8, 8)

    parameters = {
        "tbabs_1_nh": rng.uniform(0.1, 0.4, size=num_params),
        "powerlaw_1_alpha": rng.uniform(1, 3, size=num_params),
        "powerlaw_1_norm": rng.exponential(10 ** (-0.5), size=num_params),
        "blackbodyrad_1_kT": rng.uniform(0.1, 3.0, size=num_params),
        "blackbodyrad_1_norm": rng.exponential(10 ** (-3), size=num_params),
    }

    return parameters


@pytest.fixture
def unidimensional_parameters():
    from numpy.random import default_rng

    rng = default_rng(42)

    num_params = 16

    parameters = {
        "tbabs_1_nh": rng.uniform(0.1, 0.4, size=num_params),
        "powerlaw_1_alpha": rng.uniform(1, 3, size=num_params),
        "powerlaw_1_norm": rng.exponential(10 ** (-0.5), size=num_params),
        "blackbodyrad_1_kT": rng.uniform(0.1, 3.0, size=num_params),
        "blackbodyrad_1_norm": rng.exponential(10 ** (-3), size=num_params),
    }

    return parameters


@pytest.fixture
def model():
    from jaxspec.model.additive import Blackbodyrad, Powerlaw
    from jaxspec.model.multiplicative import Tbabs

    return Tbabs() * (Powerlaw() + Blackbodyrad())


@pytest.fixture
def sharded_parameters(unidimensional_parameters):
    from jax.experimental import mesh_utils
    from jax.sharding import PositionalSharding

    sharding = PositionalSharding(mesh_utils.create_device_mesh((4,)))

    return jax.device_put(unidimensional_parameters, sharding)


def test_fakeits_apply_stat(obsconfs, model, multidimensional_parameters):
    obsconf = obsconfs[0]
    spectra = fakeit_for_multiple_parameters(
        obsconf, model, multidimensional_parameters, apply_stat=False
    )
    chex.assert_type(spectra, float)

    spectra = fakeit_for_multiple_parameters(
        obsconf, model, multidimensional_parameters, apply_stat=True
    )
    chex.assert_type(spectra, int)


def test_fakeits_parallel(obsconfs, model, sharded_parameters):
    obsconf = obsconfs[0]
    spectra = fakeit_for_multiple_parameters(obsconf, model, sharded_parameters, apply_stat=False)
    chex.assert_type(spectra, float)

    spectra = fakeit_for_multiple_parameters(obsconf, model, sharded_parameters, apply_stat=True)
    chex.assert_type(spectra, int)


def test_fakeits_multiple_observation(obsconfs, model, multidimensional_parameters):
    obsconf = obsconfs[0]
    spectra = fakeit_for_multiple_parameters(
        obsconf, model, multidimensional_parameters, apply_stat=False
    )
    chex.assert_type(spectra, float)

    spectra = fakeit_for_multiple_parameters(
        obsconf, model, multidimensional_parameters, apply_stat=True
    )
    chex.assert_type(spectra, int)


def test_mock_obsconf(instruments, model, multidimensional_parameters):
    for instrument in instruments:
        obsconf = ObsConfiguration.mock_from_instrument(instrument, exposure=1e5)
        fakeit_for_multiple_parameters(obsconf, model, multidimensional_parameters)



import numpy as np
import scipy
import sparse
import xarray as xr

from .instrument import Instrument
from .observation import Observation


class ObsConfiguration(xr.Dataset):
    """
    Class to store the data of a folding model, which is the link between the unfolded and folded spectra.
    """

    transfer_matrix: xr.DataArray
    """The transfer matrix"""
    area: xr.DataArray
    """The effective area of the instrument"""
    exposure: xr.DataArray
    """The total exposure"""
    folded_counts: xr.DataArray
    """The observed counts, after grouping"""
    folded_background: xr.DataArray
    """The background counts, after grouping"""

    __slots__ = (
        "transfer_matrix",
        "area",
        "exposure",
        "folded_counts",
        "folded_background",
    )

    @property
    def in_energies(self):
        """
        The energy bounds of the unfolded bins in keV. The shape is (2, n_bins).
        """

        in_energies = np.stack(
            (
                np.asarray(self.coords["e_min_unfolded"], dtype=np.float64),
                np.asarray(self.coords["e_max_unfolded"], dtype=np.float64),
            )
        )

        return in_energies

    @property
    def out_energies(self):
        """
        The energy bounds of the folded bins in keV. The shape is (2, n_bins).
        """

        out_energies = np.stack(
            (
                np.asarray(self.coords["e_min_folded"].data, dtype=np.float64),
                np.asarray(self.coords["e_max_folded"].data, dtype=np.float64),
            )
        )

        return out_energies

    @classmethod
    def from_pha_file(
        cls,
        pha_path,
        rmf_path: str | None = None,
        arf_path: str | None = None,
        bkg_path: str | None = None,
        low_energy: float = 1e-20,
        high_energy: float = 1e20,
    ):
        r"""
        Build the observation configuration from a PHA file.

        Parameters:
            pha_path: The path to the PHA file.
            rmf_path: The path to the RMF file.
            arf_path: The path to the ARF file.
            bkg_path: The path to the background file.
            low_energy: The lower bound of the energy range to consider.
            high_energy: The upper bound of the energy range to consider.
        """

        from .util import data_path_finder

        arf_path_default, rmf_path_default, bkg_path_default = data_path_finder(
            pha_path,
            require_arf=(arf_path is None) and (arf_path != ""),
            require_rmf=rmf_path is None,
            require_bkg=bkg_path is None,
        )

        arf_path = arf_path_default if arf_path is None else arf_path
        rmf_path = rmf_path_default if rmf_path is None else rmf_path
        bkg_path = bkg_path_default if bkg_path is None else bkg_path

        instrument = Instrument.from_ogip_file(
            rmf_path, arf_path=arf_path if arf_path != "" else None
        )
        observation = Observation.from_pha_file(pha_path, bkg_path=bkg_path)

        return cls.from_instrument(
            instrument, observation, low_energy=low_energy, high_energy=high_energy
        )

    @classmethod
    def from_instrument(
        cls,
        instrument: Instrument,
        observation: Observation,
        low_energy: float = 1e-20,
        high_energy: float = 1e20,
    ):
        r"""
        Build the observation configuration from an [`Instrument`][jaxspec.data.Instrument] and an [`Observation`][jaxspec.data.Observation] object.

        Parameters:
            instrument: The instrument object.
            observation: The observation object.
            low_energy: The lower bound of the energy range to consider.
            high_energy: The upper bound of the energy range to consider.

        """
        # First we unpack all the xarray data to classical np array for efficiency
        # We also exclude the bins that are flagged with bad quality on the instrument
        quality_filter = observation.quality.data == 0
        grouping = (
            scipy.sparse.csr_array(observation.grouping.data.to_scipy_sparse()) * quality_filter
        )
        e_min_channel = instrument.coords["e_min_channel"].data
        e_max_channel = instrument.coords["e_max_channel"].data
        e_min_unfolded = instrument.coords["e_min_unfolded"].data
        e_max_unfolded = instrument.coords["e_max_unfolded"].data
        redistribution = scipy.sparse.csr_array(instrument.redistribution.data.to_scipy_sparse())
        area = instrument.area.data
        exposure = observation.exposure.data

        # Computing the lower and upper energies of the bins after grouping
        # This is just a trick to compute it without 10 lines of code
        grouping_nan = observation.grouping.data * quality_filter
        grouping_nan.fill_value = np.nan
        e_min = sparse.nanmin(grouping_nan * e_min_channel, axis=1).todense()
        e_max = sparse.nanmax(grouping_nan * e_max_channel, axis=1).todense()

        # Compute the transfer matrix
        transfer_matrix = grouping @ (redistribution * area * exposure)

        # Exclude bins out of the considered energy range, and bins without contribution from the RMF
        row_idx = (e_min > low_energy) & (e_max < high_energy) & (grouping.sum(axis=1) > 0)
        col_idx = (e_min_unfolded > 0) & (redistribution.sum(axis=0) > 0)

        # Apply this reduction to all the relevant arrays
        transfer_matrix = sparse.COO.from_scipy_sparse(transfer_matrix[row_idx][:, col_idx])
        folded_counts = observation.folded_counts.data[row_idx]
        folded_backratio = observation.folded_backratio.data[row_idx]
        area = instrument.area.data[col_idx]
        e_min_folded = e_min[row_idx]
        e_max_folded = e_max[row_idx]
        e_min_unfolded = e_min_unfolded[col_idx]
        e_max_unfolded = e_max_unfolded[col_idx]

        if observation.folded_background is not None:
            folded_background = observation.folded_background.data[row_idx]
            folded_background_unscaled = observation.folded_background_unscaled.data[row_idx]
        else:
            folded_background = np.zeros_like(folded_counts)
            folded_background_unscaled = np.zeros_like(folded_counts)

        data_dict = {
            "transfer_matrix": (
                ["folded_channel", "unfolded_channel"],
                transfer_matrix,
                {
                    "description": "Transfer matrix to use to fold the incoming spectrum. It is built and restricted using the grouping, redistribution matrix, effective area, quality flags and energy bands defined by the user."
                },
            ),
            "area": (
                ["unfolded_channel"],
                area,
                {
                    "description": "Effective area with the same restrictions as the transfer matrix.",
                    "units": "cm^2",
                },
            ),
            "exposure": ([], exposure, {"description": "Total exposure", "unit": "s"}),
            "folded_counts": (
                ["folded_channel"],
                folded_counts,
                {
                    "description": "Folded counts after grouping, with the same restrictions as the transfer matrix.",
                    "unit": "photons",
                },
            ),
            "folded_backratio": (
                ["folded_channel"],
                folded_backratio,
                {
                    "description": "Background scaling after grouping, with the same restrictions as the transfer matrix."
                },
            ),
            "folded_background": (
                ["folded_channel"],
                folded_background,
                {
                    "description": "Folded background counts after grouping, with the same restrictions as the transfer matrix.",
                    "unit": "photons",
                },
            ),
            "folded_background_unscaled": (
                ["folded_channel"],
                folded_background_unscaled,
                {
                    "description": "To be done",
                    "unit": "photons",
                },
            ),
        }

        return cls(
            data_dict,
            coords={
                "e_min_folded": (
                    ["folded_channel"],
                    e_min_folded,
                    {"description": "Low energy of folded channel"},
                ),
                "e_max_folded": (
                    ["folded_channel"],
                    e_max_folded,
                    {"description": "High energy of folded channel"},
                ),
                "e_min_unfolded": (
                    ["unfolded_channel"],
                    e_min_unfolded,
                    {"description": "Low energy of unfolded channel"},
                ),
                "e_max_unfolded": (
                    ["unfolded_channel"],
                    e_max_unfolded,
                    {"description": "High energy of unfolded channel"},
                ),
            },
            attrs=observation.attrs | instrument.attrs,
        )

    @classmethod
    def mock_from_instrument(
        cls,
        instrument: Instrument,
        exposure: float,
        low_energy: float = 1e-300,
        high_energy: float = 1e300,
    ):
        """
        Create a mock observation configuration from an instrument object. The fake observation will have zero counts.

        Parameters:
            instrument: The instrument object.
            exposure: The total exposure of the mock observation.
            low_energy: The lower bound of the energy range to consider.
            high_energy: The upper bound of the energy range to consider.
        """

        n_channels = len(instrument.coords["instrument_channel"])

        observation = Observation.from_matrix(
            np.zeros(n_channels),
            sparse.eye(n_channels),
            np.arange(n_channels),
            np.zeros(n_channels, dtype=bool),
            exposure,
            backratio=np.ones(n_channels),
            attributes={"description": "Mock observation"} | instrument.attrs,
        )

        return cls.from_instrument(
            instrument, observation, low_energy=low_energy, high_energy=high_energy
        )

    def plot_counts(self, **kwargs):
        return self.folded_counts.plot.step(x="e_min_folded", where="post", **kwargs)



import shutil

import matplotlib.pyplot as plt
import pytest

from astropy.io import fits
from conftest import data_collection, data_directory
from jaxspec.data import Instrument, ObsConfiguration, Observation


def test_loading_non_existent_files():
    path = "this/path/does/not/exist"

    with pytest.raises(FileNotFoundError):
        Instrument.from_ogip_file(path)

    with pytest.raises(FileNotFoundError):
        Observation.from_pha_file(path)

    with pytest.raises(FileNotFoundError):
        ObsConfiguration.from_pha_file(path)


@pytest.mark.parametrize("observation", data_collection, ids=lambda m: m["name"])
def test_loading_curated_data_files_from_pha(observation):
    # Not working either because the header are wrong or the file is a pha2
    not_working = ["XMM-Newton/RGS", "XRISM/Resolve", "Chandra/LETGS"]

    if observation["name"] not in not_working:
        ObsConfiguration.from_pha_file(data_directory / observation["pha_path"])


@pytest.mark.parametrize("observation", data_collection, ids=lambda m: m["name"])
def test_plot_curated_data_files_grouping(observation):
    # Not working because the file is a pha2
    not_working = ["Chandra/LETGS", "IXPE/GPD"]

    if observation["name"] not in not_working:
        obs = Observation.from_pha_file(data_directory / observation["pha_path"])

        if observation["name"] in ["XRISM/Resolve", "Hitomi/SXS"]:
            obs.plot_counts()
            plt.suptitle(observation["name"])
            plt.show()

        else:
            obs.plot_grouping()
            plt.suptitle(observation["name"])
            plt.show()


@pytest.mark.parametrize("observation", data_collection, ids=lambda m: m["name"])
def test_loading_curated_data_files_from_pha_with_explicit_files(observation):
    # Not working because the file is a pha2
    not_working = ["Chandra/LETGS"]

    if observation["name"] not in not_working:
        ObsConfiguration.from_pha_file(
            data_directory / observation["pha_path"],
            rmf_path=data_directory / observation["rmf_path"],
            arf_path=data_directory / observation.get("arf_path", None)
            if observation.get("arf_path", None) is not None
            else None,
        )


@pytest.mark.parametrize("observation", data_collection, ids=lambda m: m["name"])
def test_plot_instruments_from_curated_data_files(observation):
    title = observation["name"]

    if observation["name"] not in ["Chandra/LETGS"]:
        if "arf_path" in observation.keys():
            instrument = Instrument.from_ogip_file(
                data_directory / observation["rmf_path"],
                arf_path=data_directory / observation["arf_path"],
            )

            title += " (ARF included)"

        else:
            instrument = Instrument.from_ogip_file(
                data_directory / observation["rmf_path"], arf_path=None
            )

            title += " (no ARF included)"

        instrument.plot_area()
        plt.suptitle(title)
        plt.show()

        if observation["name"] not in ["XRISM/Resolve", "Hitomi/SXS"]:
            instrument.plot_redistribution()
            plt.suptitle(observation["name"])
            plt.show()


def test_plot_exemple_instruments(instruments: list[Instrument]):
    for instrument in instruments:
        plt.figure(figsize=(10, 5))
        plt.subplot(121)
        instrument.plot_redistribution()
        plt.subplot(122)
        instrument.plot_area()
        plt.show()


def test_plot_exemple_observations(observations: list[Observation]):
    for observation in observations:
        observation.plot_grouping()
        plt.show()


@pytest.fixture
def file_with_no_grouping(tmp_path):
    file_path = tmp_path / "no_grouping.fits"
    shutil.copyfile(data_directory / "XMM-Newton/EPIC-PN/PN_spectrum_grp20.fits", file_path)
    shutil.copyfile(data_directory / "XMM-Newton/EPIC-PN/PN.arf", tmp_path / "PN.arf")
    shutil.copyfile(data_directory / "XMM-Newton/EPIC-PN/PN.rmf", tmp_path / "PN.rmf")
    shutil.copyfile(
        data_directory / "XMM-Newton/EPIC-PN/PNbackground_spectrum.fits",
        tmp_path / "PNbackground_spectrum.fits",
    )

    with fits.open(file_path) as hdul:
        hdul[1].columns.del_col("GROUPING")
        hdul.writeto(file_path, overwrite=True)

    return file_path


def test_loading_file_with_no_grouping(file_with_no_grouping):
    with pytest.raises(ValueError):
        Observation.from_pha_file(file_with_no_grouping)


@pytest.fixture
def file_with_no_backscal(tmp_path):
    file_path = tmp_path / "no_backscal.fits"
    shutil.copyfile(data_directory / "XMM-Newton/EPIC-PN/PN_spectrum_grp20.fits", file_path)
    shutil.copyfile(data_directory / "XMM-Newton/EPIC-PN/PN.arf", tmp_path / "PN.arf")
    shutil.copyfile(data_directory / "XMM-Newton/EPIC-PN/PN.rmf", tmp_path / "PN.rmf")
    shutil.copyfile(
        data_directory / "XMM-Newton/EPIC-PN/PNbackground_spectrum.fits",
        tmp_path / "PNbackground_spectrum.fits",
    )

    with fits.open(file_path) as hdul:
        del hdul[1].header["BACKSCAL"]
        hdul.writeto(file_path, overwrite=True)

    return file_path


def test_loading_file_with_no_backscal(file_with_no_backscal):
    with pytest.raises(ValueError):
        Observation.from_pha_file(file_with_no_backscal)


@pytest.fixture
def file_with_no_areascal(tmp_path):
    file_path = tmp_path / "no_areascal.fits"
    shutil.copyfile(data_directory / "XMM-Newton/EPIC-PN/PN_spectrum_grp20.fits", file_path)
    shutil.copyfile(data_directory / "XMM-Newton/EPIC-PN/PN.arf", tmp_path / "PN.arf")
    shutil.copyfile(data_directory / "XMM-Newton/EPIC-PN/PN.rmf", tmp_path / "PN.rmf")
    shutil.copyfile(
        data_directory / "XMM-Newton/EPIC-PN/PNbackground_spectrum.fits",
        tmp_path / "PNbackground_spectrum.fits",
    )

    with fits.open(file_path) as hdul:
        del hdul[1].header["AREASCAL"]
        hdul.writeto(file_path, overwrite=True)

    return file_path


def test_loading_file_with_no_areascal(file_with_no_areascal):
    with pytest.raises(ValueError):
        Observation.from_pha_file(file_with_no_areascal)



from jaxspec.util.misc import catchtime


def test_func():
    return 1


def test_catchtime_measures_time_correctly():
    with catchtime(desc="Test Task", print_time=False) as get_time:
        result = test_func()
    assert result == 1
    assert get_time() > 0


def test_catchtime_prints_time_when_print_time_is_true(capfd):
    with catchtime(desc="Test Task", print_time=True) as get_time:
        result = test_func()
    out, err = capfd.readouterr()
    assert "Test Task" in out
    assert "seconds" in out


def test_catchtime_does_not_print_time_when_print_time_is_false(capfd):
    with catchtime(desc="Test Task", print_time=False) as get_time:
        result = test_func()
    out, err = capfd.readouterr()
    assert out == ""



r'''
import os
import re

from dataclasses import dataclass

import jax.numpy as jnp
import matplotlib.pyplot as plt
import numpy as np
import pytest

from jax.tree_util import tree_map
from jaxspec.data import ObsConfiguration
from jaxspec.data.util import fakeit_for_multiple_parameters
from jaxspec.model.abc import SpectralModel
from jaxspec.util.online_storage import table_manager
from scipy.stats import anderson_ksamp

xspec = pytest.importorskip("xspec")


@dataclass
class AdditiveModelTestSetup:
    """Class to setup a test case for a spectral model"""

    name_xspec: str  # Name of the model in xspec
    name_jaxspec: str  # Name of the model in jaxspec
    parameters: dict[str, float]  # Parameter values
    parameters_order: list[str]  # Parameter order in xspec
    required_tol: float = 1e-2  # Required tolerance between xspec and jaxspec
    weight: float = 1  # Weight of the XSPEC model when systematic biases are present (e.g. add an epsilon to the model)
    energy_range: tuple[float, float, int] = (
        0.2,
        20,
        10000,
    )  # Energy range to test, lower, upper, number of bins
    model_string: list[tuple[str, str]] | None = None  # Model string to use in XSPEC


models_to_test = [
    AdditiveModelTestSetup(
        name_xspec="powerlaw",
        name_jaxspec="Powerlaw()",
        parameters={"alpha": 1.5, "norm": 1},
        parameters_order=["alpha", "norm"],
    ),
    AdditiveModelTestSetup(
        name_xspec="bbody",
        name_jaxspec="Blackbody()",
        parameters={"kT": 3, "norm": 1},
        parameters_order=["kT", "norm"],
    ),
    AdditiveModelTestSetup(
        name_xspec="bbodyrad",
        name_jaxspec="Blackbodyrad()",
        parameters={"kT": 3, "norm": 1},
        parameters_order=["kT", "norm"],
    ),
    AdditiveModelTestSetup(
        name_xspec="lorentz",
        name_jaxspec="Lorentz()",
        parameters={"E_l": 3, "sigma": 1, "norm": 1},
        parameters_order=["E_l", "sigma", "norm"],
        weight=1
        / 1.02475,  # There is an epsilon in XSPEC to avoid division by zero which add a 2% bias
        required_tol=2e-1,  # The bias is too large to be corrected by the weight above, I shall find why
        energy_range=(4, 9, 10000),
    ),
    AdditiveModelTestSetup(
        name_xspec="gauss",
        name_jaxspec="Gauss()",
        parameters={"E_l": 3, "sigma": 1, "norm": 1},
        parameters_order=["E_l", "sigma", "norm"],
        energy_range=(4, 9, 10000),
    ),
    AdditiveModelTestSetup(
        name_xspec="cutoffpl",
        name_jaxspec="Cutoffpl()",
        parameters={"alpha": 1, "beta": 15, "norm": 1},
        parameters_order=["alpha", "beta", "norm"],
    ),
]


def plot_comparison(e_bins, spec_xspec, spec_jaxspec):
    fig, axs = plt.subplots(2, 1, figsize=(6, 6), sharex=True, height_ratios=[0.7, 0.3])
    e_list = e_bins.mean(axis=0)
    axs[0].plot(e_list, spec_xspec, label="xspec")
    axs[0].plot(e_list, spec_jaxspec, label="jaxspec")
    axs[0].loglog()
    axs[1].hexbin(
        e_list, (spec_xspec - spec_jaxspec) / spec_xspec, gridsize=100, xscale="log", cmap="Blues"
    )
    axs[0].legend()
    axs[1].set_xlabel("Energy (keV)")
    axs[1].set_ylabel("(xspec-jaxspec)/xspec")
    plt.show()


@pytest.mark.parametrize(
    "model", models_to_test, ids=lambda m: f"{m.name_xspec} & {m.name_jaxspec}"
)
def test_models(model: AdditiveModelTestSetup):
    xspec.AllModels.clear()
    xspec.AllData.clear()
    xspec.AllData.dummyrsp(*model.energy_range, "lin")

    m_xspec = xspec.Model(model.name_xspec)

    if model.model_string is not None:
        for model_string in model.model_string:
            xspec.Xset.addModelString(*model_string)

    xspec_pars = []
    for par in model.parameters_order:
        xspec_pars.append(par)

    m_xspec.setPars(*[model.parameters[par] for par in model.parameters_order])
    m_xspec.show()
    xspec_energies = np.array(m_xspec.energies(0))
    e_bins = np.vstack([xspec_energies[:-1], xspec_energies[1:]])
    spec_xspec = np.array(m_xspec.values(0)) * model.weight

    m_jaxspec = SpectralModel.from_string(model.name_jaxspec)
    p_jaxspec = {re.sub(r"\(.*\)", "", model.name_jaxspec.lower()) + r"_1": model.parameters}

    spec_jaxspec = m_jaxspec.photon_flux(tree_map(jnp.asarray, p_jaxspec), *e_bins, n_points=20)

    assert np.isclose(
        spec_jaxspec, spec_xspec, rtol=model.required_tol
    ).all(), f"jaxspec {model.name_jaxspec} and xspec {model.name_xspec} are not close enough"


@pytest.mark.filterwarnings("ignore:p-value capped")
@pytest.mark.parametrize(
    "model", models_to_test, ids=lambda m: f"{m.name_xspec} & {m.name_jaxspec}"
)
def test_fakeits(
    tmp_path, request, monkeypatch, model: AdditiveModelTestSetup, exposure=float(10_000)
):
    pn_rmf_path = table_manager.fetch("example_data/NGC7793_ULX4/PN.rmf")
    pn_arf_path = table_manager.fetch("example_data/NGC7793_ULX4/PN.arf")
    dir_path = os.path.dirname(pn_rmf_path)
    monkeypatch.chdir(os.path.join(os.path.dirname(request.fspath.dirname), dir_path))

    # XSPEC's fakeit
    xspec.AllModels.clear()
    xspec.AllData.clear()
    xspec.AllData.dummyrsp(*model.energy_range, "lin")

    n_spectra = 100

    m_xspec = xspec.Model(model.name_xspec)

    fakeit_settings = n_spectra * [
        xspec.FakeitSettings(
            response="PN.rmf", arf="PN.arf", exposure=exposure, fileName="fakeit.pha"
        )
    ]

    xspec_pars = []
    for par in model.parameters_order:
        xspec_pars.append(par)

    m_xspec.setPars(*[model.parameters[par] for par in model.parameters_order])
    xspec.AllData.fakeit(
        nSpectra=n_spectra, settings=fakeit_settings, applyStats=True, filePrefix=""
    )
    xspec.AllData.ignore(f"0.0-{model.energy_range[0]:.1f} {model.energy_range[1]:.1f}-**")

    xspec_spectra = np.empty((n_spectra, len(xspec.AllData(1).values)))

    for i in range(1, n_spectra + 1):
        xspec_spectra[i - 1] = (np.asarray(xspec.AllData(i).values) * exposure).astype(int)

    # JAXSPEC's fakeit
    m_jaxspec = SpectralModel.from_string(model.name_jaxspec)
    p_jaxspec = {re.sub(r"\(.*\)", "", model.name_jaxspec.lower()) + r"_1": model.parameters}
    p_jaxspec = tree_map(lambda x: x * jnp.ones(n_spectra), p_jaxspec)

    setup = ObsConfiguration.from_pha_file(
        "fakeit.pha", low_energy=model.energy_range[0], high_energy=model.energy_range[1]
    )

    jaxspec_spectra = fakeit_for_multiple_parameters(setup, m_jaxspec, p_jaxspec)

    assert anderson_ksamp(np.hstack((xspec_spectra, jaxspec_spectra))).pvalue >= 0.25
'''



from abc import ABC, abstractmethod
from collections.abc import Callable

import numpyro

from flax import nnx
from numpyro.distributions import Distribution


class GainModel(ABC, nnx.Module):
    @abstractmethod
    def numpyro_model(self, observation_name: str):
        pass


class ConstantGain(GainModel):
    def __init__(self, prior_distribution: Distribution):
        self.prior_distribution = prior_distribution

    def numpyro_model(self, observation_name: str):
        factor = numpyro.sample(f"ins/~/gain_{observation_name}", self.prior_distribution)

        def gain(energy):
            return factor

        return gain


class ShiftModel(ABC, nnx.Module):
    @abstractmethod
    def numpyro_model(self, observation_name: str):
        pass


class ConstantShift(ShiftModel):
    def __init__(self, prior_distribution: Distribution):
        self.prior_distribution = prior_distribution

    def numpyro_model(self, observation_name: str):
        shift_offset = numpyro.sample(f"ins/~/shift_{observation_name}", self.prior_distribution)

        def shift(energy):
            return energy + shift_offset

        return shift


class InstrumentModel(nnx.Module):
    def __init__(
        self,
        reference_observation_name: str,
        gain_model: GainModel | None = None,
        shift_model: ShiftModel | None = None,
    ):
        self.reference = reference_observation_name
        self.gain_model = gain_model
        self.shift_model = shift_model

    def get_gain_and_shift_model(
        self, observation_name: str
    ) -> tuple[Callable | None, Callable | None]:
        """
        Return the gain and shift models for the given observation. It should be called within a numpyro model.
        """

        if observation_name == self.reference:
            return None, None

        else:
            gain = (
                self.gain_model.numpyro_model(observation_name)
                if self.gain_model is not None
                else None
            )
            shift = (
                self.shift_model.numpyro_model(observation_name)
                if self.shift_model is not None
                else None
            )

            return gain, shift



"""Helper functions to deal with the graph logic within model building"""

import re

from collections.abc import Callable
from uuid import uuid4

import networkx as nx


def get_component_names(graph: nx.DiGraph):
    """
    Get the set of component names from the nodes of a graph.

    Parameters:
        graph: The graph to get the component names from.
    """
    return set(
        data["name"] for _, data in graph.nodes(data=True) if "component" in data.get("type")
    )


def increment_name(name: str, used_names: set):
    """
    Increment the suffix number in a name if it is formated as 'name_1'.

    Parameters:
        name: The name to increment.
        used_names: The set of names that are already used.
    """
    # Use regex to extract base name and suffix number
    match = re.match(r"(.*?)(?:_(\d+))?$", name)
    base_name = match.group(1)
    suffix = match.group(2)
    if suffix:
        number = int(suffix)
    else:
        number = 1  # Start from 1 if there is no suffix

    new_name = name
    while new_name in used_names:
        number += 1
        new_name = f"{base_name}_{number}"

    return new_name


def compose_with_rename(graph_1: nx.DiGraph, graph_2: nx.DiGraph):
    """
    Compose two graphs by updating the 'name' attributes of nodes in graph_2,
    and return the graph joined on the 'out' node.

    Parameters:
        graph_1: The first graph to compose.
        graph_2: The second graph to compose.
    """

    # Initialize the set of used names with names from graph_1
    used_names = get_component_names(graph_1)

    # Update the 'name' attributes in graph_2 to make them unique
    for node, data in graph_2.nodes(data=True):
        if "component" in data.get("type"):
            original_name = data["name"]
            new_name = original_name

            if new_name in used_names:
                new_name = increment_name(original_name, used_names)
                data["name"] = new_name
                used_names.add(new_name)

            else:
                used_names.add(new_name)

    # Now you can safely compose the graphs
    composed_graph = nx.compose(graph_1, graph_2)

    return composed_graph


def compose(
    graph_1: nx.DiGraph,
    graph_2: nx.DiGraph,
    operation: str = "",
    operation_func: Callable = lambda x, y: None,
):
    """
    Compose two graphs by joining the 'out' node of graph_1 and graph_2, and turning
    it to an 'operation' node with the relevant operator and add a new 'out' node.

    Parameters:
        graph_1: The first graph to compose.
        graph_2: The second graph to compose.
        operation: The string describing the operation to perform.
        operation_func: The callable that performs the operation.
    """

    combined_graph = compose_with_rename(graph_1, graph_2)
    node_id = str(uuid4())
    graph = nx.relabel_nodes(combined_graph, {"out": node_id})
    nx.set_node_attributes(graph, {node_id: f"{operation}_operation"}, "type")
    nx.set_node_attributes(graph, {node_id: operation_func}, "operator")

    # Now add the output node and link it to the operation node
    graph.add_node("out", type="out")
    graph.add_edge(node_id, "out")

    # Compute the new depth of each node
    longest_path = nx.dag_longest_path_length(graph)

    for node in graph.nodes:
        nx.set_node_attributes(
            graph,
            {node: longest_path - nx.shortest_path_length(graph, node, "out")},
            "depth",
        )

    return graph


def export_to_mermaid(graph, file=None):
    mermaid_code = "graph LR\n"  # LR = left to right

    # Add nodes
    for node, attributes in graph.nodes(data=True):
        if attributes["type"] == "out":
            mermaid_code += f'    {node}("Output")\n'

        else:
            operation_type, node_type = attributes["type"].split("_")

            if node_type == "component":
                name, number = attributes["name"].split("_")
                mermaid_code += f'    {node}("{name.capitalize()} ({number})")\n'

            elif node_type == "operation":
                if operation_type == "add":
                    mermaid_code += f"    {node}{{**+**}}\n"

                elif operation_type == "mul":
                    mermaid_code += f"    {node}{{**x**}}\n"

    # Draw connexion between nodes
    for source, target in graph.edges():
        mermaid_code += f"    {source} --> {target}\n"

    if file is None:
        return mermaid_code
    else:
        with open(file, "w") as f:
            f.write(mermaid_code)



import astropy.units as u

from .instrument import Instrument
from .obsconf import ObsConfiguration
from .observation import Observation

u.add_enabled_aliases({"counts": u.count})
u.add_enabled_aliases({"channel": u.dimensionless_unscaled})
# Arbitrary units are found in .rsp files , let's hope it is compatible with what we would expect as the rmf x arf
# u.add_enabled_aliases({"au": u.dimensionless_unscaled})



from watermark import watermark


def debug_info():
    """
    Script which displays useful information about the user system and environment
    """

    # Cimer CamilleTheBest pour l'idée
    print(watermark())
    print(watermark(packages="jaxspec,jax,jaxlib,numpyro,flax,numpy,scipy"))



from collections.abc import Callable
from contextlib import contextmanager
from time import perf_counter


@contextmanager
def catchtime(desc="Task", print_time=True) -> Callable[[], float]:
    """
    Context manager to measure time taken by a task.

    Parameters
    ----------
        desc (str): Description of the task.
        print_time (bool): Whether to print the time taken by the task.

    Returns:
    -------
        Callable[[], float]: Function to get the time taken by the task.
    """

    t1 = t2 = perf_counter()
    yield lambda: t2 - t1
    t2 = perf_counter()
    if print_time:
        print(f"{desc}: {t2 - t1:.3f} seconds")



import jax

from jax import random
from numpyro.contrib.nested_sampling import NestedSampler

from ..analysis.results import FitResult
from ..fit._fitter import BayesianModelFitter


class NSFitter(BayesianModelFitter):
    r"""
    A class to fit a model to a given set of observation using the Nested Sampling algorithm. This class uses the
    [`DefaultNestedSampler`][jaxns.DefaultNestedSampler] from [`jaxns`](https://jaxns.readthedocs.io/en/latest/) which
    implements the [Phantom-Powered Nested Sampling](https://arxiv.org/abs/2312.11330) algorithm.

    !!! info
        Ensure large prior volume is covered by the prior distributions to ensure the algorithm yield proper results.

    """

    def fit(
        self,
        rng_key: int = 0,
        num_samples: int = 1000,
        num_live_points: int = 1000,
        plot_diagnostics=False,
        termination_kwargs: dict | None = None,
        verbose=True,
        use_transformed_model: bool = True,
    ) -> FitResult:
        """
        Fit the model to the data using the Phantom-Powered nested sampling algorithm.

        Parameters:
            rng_key: the random key used to initialize the sampler.
            num_samples: the number of samples to draw.
            num_live_points: the number of live points to use at the start of the NS algorithm.
            plot_diagnostics: whether to plot the diagnostics of the NS algorithm.
            termination_kwargs: additional arguments to pass to the termination criterion of the NS algorithm.
            verbose: whether to print the progress of the NS algorithm.

        Returns:
            A [`FitResult`][jaxspec.analysis.results.FitResult] instance containing the results of the fit.
        """

        bayesian_model = self.transformed_numpyro_model
        keys = random.split(random.PRNGKey(rng_key), 4)

        ns = NestedSampler(
            bayesian_model,
            constructor_kwargs=dict(
                verbose=verbose,
                difficult_model=True,
                max_samples=1e5,
                parameter_estimation=True,
                gradient_guided=True,
                devices=jax.devices(),
                # init_efficiency_threshold=0.01,
                num_live_points=num_live_points,
            ),
            termination_kwargs=termination_kwargs if termination_kwargs else dict(),
        )

        ns.run(keys[0])

        if plot_diagnostics:
            ns.diagnostics()

        posterior = ns.get_samples(keys[1], num_samples=num_samples)
        inference_data = self.build_inference_data(
            posterior, num_chains=1, use_transformed_model=use_transformed_model
        )

        return FitResult(
            self,
            inference_data,
            background_model=self.background_model,
        )



from contextlib import nullcontext as does_not_raise

import numpyro.distributions as dist
import pytest

from conftest import (
    dict_of_obsconf,
    list_of_obsconf,
    prior_shared_pars,
    prior_split_pars,
    single_obsconf,
    spectral_model,
)
from jaxspec.fit import BayesianModel, MCMCFitter, TiedParameter, VIFitter
from jaxspec.model.additive import Powerlaw
from jaxspec.model.instrument import ConstantGain, ConstantShift, InstrumentModel
from jaxspec.model.multiplicative import Tbabs
from numpyro.optim import optax_to_numpyro
from optax import adamw

sparsify_marker = pytest.mark.parametrize(
    "sparse",
    [
        pytest.param(True, id="sparse matrix"),
        pytest.param(False, id="dense matrix"),
    ],
)


data_prior_marker = pytest.mark.parametrize(
    "model, prior, obsconf, expectation",
    [
        pytest.param(
            spectral_model,
            prior_shared_pars,
            single_obsconf,
            does_not_raise(),
            id="single observation-shared parameters",
        ),
        pytest.param(
            spectral_model,
            prior_split_pars,
            single_obsconf,
            pytest.raises(ValueError),
            id="single observation-split parameters",
        ),
        pytest.param(
            spectral_model,
            prior_shared_pars,
            list_of_obsconf,
            does_not_raise(),
            id="list of observation-shared parameters",
        ),
        pytest.param(
            spectral_model,
            prior_shared_pars,
            dict_of_obsconf,
            does_not_raise(),
            id="dict of observation-shared parameters",
        ),
        pytest.param(
            spectral_model,
            prior_split_pars,
            list_of_obsconf,
            does_not_raise(),
            id="list of observation-split parameters",
        ),
        pytest.param(
            spectral_model,
            prior_split_pars,
            dict_of_obsconf,
            does_not_raise(),
            id="dict of observation-split parameters",
        ),
    ],
)

mcmc_marker = pytest.mark.parametrize(
    "sampler",
    [
        pytest.param("nuts", id="NUTS"),
        pytest.param("aies", id="AIES"),
        pytest.param("ess", id="ESS"),
    ],
)


@pytest.mark.fast
@sparsify_marker
@data_prior_marker
def test_build_model(model, prior, obsconf, expectation, sparse):
    """Try to build a model from the given combination of observation and priors"""
    with expectation:
        BayesianModel(model, prior, obsconf, sparsify_matrix=sparse)


@pytest.mark.slow
@sparsify_marker
@data_prior_marker
def test_mock_obs(model, prior, obsconf, expectation, sparse):
    """Try to generate mock observations from the given combination of observation and priors"""
    with expectation:
        bayesian_model = BayesianModel(model, prior, obsconf, sparsify_matrix=sparse)
        bayesian_model.mock_observations(bayesian_model.prior_samples())


@pytest.mark.slow
@mcmc_marker
@data_prior_marker
def test_run_mcmc(model, prior, obsconf, expectation, sampler):
    """Try to generate mock observations from the given combination of observation and priors"""
    with expectation:
        forward_model = MCMCFitter(model, prior, obsconf, background_model=None)
        result = forward_model.fit(
            num_chains=4,
            num_warmup=10,
            num_samples=10,
            sampler=sampler,
            mcmc_kwargs={"progress_bar": False},
        )

        result.photon_flux(0.7, 1.2, register=True)
        result.energy_flux(0.7, 1.2, register=True)
        result.luminosity(0.7, 1.2, redshift=0.01, register=True)
        [result._ppc_folded_branches(obs_id) for obs_id in result.obsconfs.keys()]
        result.to_chain("test")


@pytest.mark.slow
@data_prior_marker
def test_run_vi(model, prior, obsconf, expectation):
    """Try to generate mock observations from the given combination of observation and priors"""
    with expectation:
        forward_model = VIFitter(model, prior, obsconf, background_model=None)
        optim = optax_to_numpyro(adamw(3e-4))

        result = forward_model.fit(
            num_steps=100,
            num_samples=10,
            optimizer=optim,
            plot_diagnostics=True,
        )

        result.photon_flux(0.7, 1.2, register=True)
        result.energy_flux(0.7, 1.2, register=True)
        result.luminosity(0.7, 1.2, redshift=0.01, register=True)
        [result._ppc_folded_branches(obs_id) for obs_id in result.obsconfs.keys()]
        result.to_chain("test")


@pytest.mark.slow
@mcmc_marker
def test_instrument_model_building(sampler):
    forward_model = MCMCFitter(
        spectral_model,
        prior_shared_pars,
        dict_of_obsconf,
        background_model=None,
        instrument_model=InstrumentModel(
            "PN",
            gain_model=ConstantGain(dist.Uniform(0.8, 1.2)),
            shift_model=ConstantShift(dist.Uniform(-0.1, +0.1)),
        ),
    )

    result = forward_model.fit(
        num_chains=4,
        num_warmup=10,
        num_samples=10,
        sampler=sampler,
        mcmc_kwargs={"progress_bar": False},
    )

    result.photon_flux(0.7, 1.2, register=True)
    result.energy_flux(0.7, 1.2, register=True)
    result.luminosity(0.7, 1.2, redshift=0.01, register=True)
    [result._ppc_folded_branches(obs_id) for obs_id in result.obsconfs.keys()]
    result.to_chain("test")


@pytest.mark.slow
@mcmc_marker
def test_tied_parameters(sampler):
    spectral_model = Tbabs() * (Powerlaw() + Powerlaw())
    prior = {
        "powerlaw_1_alpha": dist.Uniform(0, 5),
        "powerlaw_1_norm": dist.LogUniform(1e-5, 1e-2),
        "powerlaw_2_alpha": TiedParameter("powerlaw_1_alpha", lambda x: 0.5 * x),
        "powerlaw_2_norm": dist.LogUniform(1e-5, 1e-2),
        "tbabs_1_nh": 0.6,
    }

    forward_model = MCMCFitter(spectral_model, prior, dict_of_obsconf)

    result = forward_model.fit(
        num_chains=4,
        num_warmup=10,
        num_samples=10,
        sampler=sampler,
        mcmc_kwargs={"progress_bar": False},
    )

    result.photon_flux(0.7, 1.2, register=True)
    result.energy_flux(0.7, 1.2, register=True)
    result.luminosity(0.7, 1.2, redshift=0.01, register=True)
    [result._ppc_folded_branches(obs_id) for obs_id in result.obsconfs.keys()]
    result.to_chain("test")



import inspect

from dataclasses import dataclass

import chex
import jax
import jax.numpy as jnp
import numpyro
import pytest

from jax.scipy.special import gamma, hyp1f1
from jax.typing import ArrayLike
from jaxspec.util.integrate import integrate_interval, integrate_positive

numpyro.enable_x64()
chex.set_n_cpu_devices(n=4)


@dataclass
class IntervalTestSetup:
    """Class to test integrating over an interval"""

    func: callable  # Function to integrate
    interval: tuple[float, float]  # Interval to integrate over
    result: ArrayLike  # Expected result


@dataclass
class PositiveTestSetup:
    """Class to test integrating over the positive real line"""

    func: callable  # Function to integrate
    result: ArrayLike  # Expected result


intervals_to_test = [
    IntervalTestSetup(lambda x: 1 / (1 + x), (0, 1), jnp.log(2)),
    IntervalTestSetup(lambda x: 1 / (1 + x**2), (0, 1), jnp.pi / 4),
    IntervalTestSetup(lambda x: x ** (-x), (0, 1), 1.291285997),
    IntervalTestSetup(lambda x: jnp.sin(x), (0, 2 * jnp.pi), 0),
]


positives_to_test = [
    PositiveTestSetup(lambda x: 1 / (1 + x) ** 2, 1),
    PositiveTestSetup(lambda x: 1 / (1 + x**2), jnp.pi / 2),
    PositiveTestSetup(lambda x: jnp.exp(-(x**2)), jnp.sqrt(jnp.pi) / 2),
    PositiveTestSetup(lambda x: jnp.sqrt(x) * jnp.exp(-x), jnp.sqrt(jnp.pi) / 2),
]


@pytest.mark.parametrize("setup", intervals_to_test)
def test_integrate_interval(setup: IntervalTestSetup):
    """
    Test integrating over an interval
    """

    assert jnp.isclose(
        integrate_interval(setup.func)(*setup.interval), setup.result
    ), inspect.getsource(setup.func)


@pytest.mark.parametrize("setup", positives_to_test)
def test_integrate_positive(setup: PositiveTestSetup):
    """
    Test integrating over the positive real line
    """

    assert jnp.isclose(integrate_positive(setup.func)(), setup.result), inspect.getsource(
        setup.func
    )


def test_integrate_interval_gradient():
    """
    Test the custom gradient using the integral definition of Kummer's hypergeometric function
    See https://functions.wolfram.com/HypergeometricFunctions/Hypergeometric1F1/
    """

    def hyp1f1_integral(a, b, z):
        def integrand(x, a, b, z):
            return jnp.exp(z * x) * x ** (a - 1.0) * (1.0 - x) ** (-a + b - 1.0)

        return (
            integrate_interval(integrand)(0.0, 1.0, a, b, z) * gamma(b) / (gamma(a) * gamma(b - a))
        )

    a = jnp.asarray(1.5)
    b = jnp.asarray(10.0)
    z = jnp.asarray(0.5)

    assert jnp.isclose(jax.grad(hyp1f1_integral)(a, b, z), jax.grad(hyp1f1)(a, b, z))


def test_integrate_positive_gradient():
    """
    Test the custom gradient using the integral definition of the gamma function
    """

    gamma_integral = integrate_positive(lambda t, z: t ** (z - 1) * jnp.exp(-t))
    z = jnp.asarray(2.5)

    assert jnp.isclose(jax.grad(gamma_integral)(z), jax.grad(gamma)(z))



import os

import numpy as np
import xarray as xr

from matplotlib import colors

from .ogip import DataARF, DataRMF


class Instrument(xr.Dataset):
    """
    Class to store the data of an instrument
    """

    redistribution: xr.DataArray
    """The photon redistribution probability matrix"""
    area: xr.DataArray
    """The effective area of the instrument"""

    __slots__ = (
        "redistribution",
        "area",
        "e_min_out",
        "e_max_out",
        "e_min_in",
        "e_max_in",
    )

    @classmethod
    def from_matrix(
        cls,
        redistribution_matrix,
        spectral_response,
        e_min_unfolded,
        e_max_unfolded,
        e_min_channel,
        e_max_channel,
    ):
        return cls(
            {
                "redistribution": (
                    ["instrument_channel", "unfolded_channel"],
                    redistribution_matrix,
                    {"description": "Redistribution matrix"},
                ),
                "area": (
                    ["unfolded_channel"],
                    np.array(spectral_response, dtype=np.float64),
                    {"description": "Effective area", "units": "cm^2"},
                ),
            },
            coords={
                "e_min_unfolded": (
                    ["unfolded_channel"],
                    np.array(e_min_unfolded, dtype=np.float64),
                    {"description": "Low bin energy for ingoing channels", "units": "keV"},
                ),
                "e_max_unfolded": (
                    ["unfolded_channel"],
                    np.array(e_max_unfolded, dtype=np.float64),
                    {"description": "High bin energy for ingoing channels", "units": "keV"},
                ),
                "e_min_channel": (
                    ["instrument_channel"],
                    np.array(e_min_channel, dtype=np.float64),
                    {"description": "Low bin energy for outgoing channels", "units": "keV"},
                ),
                "e_max_channel": (
                    ["instrument_channel"],
                    np.array(e_max_channel, dtype=np.float64),
                    {"description": "High bin energy for outgoing channels", "units": "keV"},
                ),
            },
            attrs={"description": "X-ray instrument response dataset"},
        )

    @classmethod
    def from_ogip_file(cls, rmf_path: str | os.PathLike, arf_path: str | os.PathLike | None = None):
        """
        Load the data from OGIP files.

        Parameters:
            rmf_path: The RMF file path.
            arf_path: The ARF file path.
        """

        rmf = DataRMF.from_file(rmf_path)

        if arf_path is not None:
            specresp = DataARF.from_file(arf_path).specresp

        else:
            specresp = rmf.matrix.sum(axis=0)
            rmf.sparse_matrix /= specresp

        return cls.from_matrix(
            rmf.sparse_matrix, specresp, rmf.energ_lo, rmf.energ_hi, rmf.e_min, rmf.e_max
        )

    def plot_redistribution(
        self,
        xscale: str = "log",
        yscale: str = "log",
        cmap=None,
        vmin: float = 1e-6,
        vmax: float = 1e0,
        add_labels: bool = True,
        **kwargs,
    ):
        """
        Plot the redistribution probability matrix

        Parameters:
            xscale : The scale of the x-axis.
            yscale : The scale of the y-axis.
            cmap : The colormap to use.
            vmin : The minimum value for the colormap.
            vmax : The maximum value for the colormap.
            add_labels : Whether to add labels to the plot.
            **kwargs : `kwargs` passed to https://docs.xarray.dev/en/latest/generated/xarray.plot.pcolormesh.html#xarray.plot.pcolormesh
        """

        import cmasher as cmr

        return xr.plot.pcolormesh(
            self.redistribution,
            x="e_max_unfolded",
            y="e_max_channel",
            xscale=xscale,
            yscale=yscale,
            cmap=cmr.ember_r if cmap is None else cmap,
            norm=colors.LogNorm(vmin=vmin, vmax=vmax),
            add_labels=add_labels,
            **kwargs,
        )

    def plot_area(self, xscale: str = "log", yscale: str = "log", where: str = "post", **kwargs):
        """
        Plot the effective area

        Parameters:
            xscale : The scale of the x-axis.
            yscale : The scale of the y-axis.
            where : The position of the steps.
            **kwargs : `kwargs` passed to https://docs.xarray.dev/en/latest/generated/xarray.DataArray.plot.line.html#xarray.DataArray.plot.line
        """

        return self.area.plot.step(
            x="e_min_unfolded", xscale=xscale, yscale=yscale, where=where, **kwargs
        )



from __future__ import annotations

import flax.nnx as nnx
import jax.numpy as jnp
import jax.scipy.special as jsp
import numpy as np

from astropy.table import Table

from ..util.online_storage import table_manager
from .abc import MultiplicativeComponent


class MultiplicativeConstant(MultiplicativeComponent):
    r"""
    A multiplicative constant

    !!! abstract "Parameters"
        * $K$ (`norm`) $\left[\text{dimensionless}\right]$: The multiplicative constant.

    """

    def __init__(self):
        self.norm = nnx.Param(1.0)

    def factor(self, energy):
        return self.norm


class Expfac(MultiplicativeComponent):
    r"""
    An exponential modification of a spectrum.

    $$
    \mathcal{M}(E) = \begin{cases}1 + A \exp \left(-fE\right) &
    \text{if $E>E_c$}\\1 & \text{if $E<E_c$}\end{cases}
    $$

    !!! abstract "Parameters"
        * $A$ (`A`) $\left[\text{dimensionless}\right]$ : Amplitude of the modification
        * $f$ (`f`) $\left[\text{keV}^{-1}\right]$ : Exponential factor
        * $E_c$ (`E_c`) $\left[\text{keV}\right]$: Start energy of modification

    """

    def __init__(self):
        self.A = nnx.Param(1.0)
        self.f = nnx.Param(1.0)
        self.E_c = nnx.Param(1.0)

    def factor(self, energy):
        return jnp.where(energy >= self.E_c, 1.0 + self.A * jnp.exp(-self.f * energy), 1.0)


class Tbabs(MultiplicativeComponent):
    r"""
    The Tuebingen-Boulder ISM absorption model. This model calculates the cross section for X-ray absorption by the ISM
    as the sum of the cross sections for X-ray absorption due to the gas-phase ISM, the grain-phase ISM,
    and the molecules in the ISM.

    $$
        \mathcal{M}(E) = \exp^{-N_{\text{H}}\sigma(E)}
    $$

    !!! abstract "Parameters"
        * $N_{\text{H}}$ (`nh`) $\left[\frac{\text{atoms}~10^{22}}{\text{cm}^2}\right]$ : Equivalent hydrogen column density


    !!! note
        Abundances and cross-sections $\sigma$ can be found in Wilms et al. (2000).

    """

    def __init__(self):
        table = Table.read(table_manager.fetch("xsect_tbabs_wilm.fits"))
        self._energy = np.asarray(table["ENERGY"], dtype=np.float64)
        self._sigma = np.asarray(table["SIGMA"], dtype=np.float64)
        self.nh = nnx.Param(1.0)

    def factor(self, energy):
        sigma = jnp.exp(
            jnp.interp(
                jnp.log(energy),
                jnp.log(self._energy),
                jnp.log(self._sigma),
                left="extrapolate",
                right="extrapolate",
            )
        )

        return jnp.exp(-self.nh * sigma)


class Phabs(MultiplicativeComponent):
    r"""
    A photoelectric absorption model.

    !!! abstract "Parameters"
        * $N_{\text{H}}$ (`nh`) $\left[\frac{\text{atoms}~10^{22}}{\text{cm}^2}\right]$ : Equivalent hydrogen column density

    """

    def __init__(self):
        table = Table.read(table_manager.fetch("xsect_phabs_aspl.fits"))
        self.energy = nnx.Variable(np.asarray(table["ENERGY"], dtype=np.float64))
        self.sigma = nnx.Variable(np.asarray(table["SIGMA"], dtype=np.float64))
        self.nh = nnx.Param(1.0)

    def factor(self, energy):
        sigma = jnp.interp(energy, self.energy, self.sigma, left=1e9, right=0.0)

        return jnp.exp(-self.nh * sigma)


class Wabs(MultiplicativeComponent):
    r"""
    A photo-electric absorption using Wisconsin (Morrison & McCammon 1983) cross-sections.

    ??? abstract "Parameters"
        * $N_{\text{H}}$ (`nh`) $\left[\frac{\text{atoms}~10^{22}}{\text{cm}^2}\right]$ : Equivalent hydrogen column density
    """

    def __init__(self):
        table = Table.read(table_manager.fetch("xsect_wabs_angr.fits"))
        self.energy = nnx.Variable(np.asarray(table["ENERGY"], dtype=np.float64))
        self.sigma = nnx.Variable(np.asarray(table["SIGMA"], dtype=np.float64))
        self.nh = nnx.Param(1.0)

    def factor(self, energy):
        sigma = jnp.interp(energy, self.energy, self.sigma, left=1e9, right=0.0)

        return jnp.exp(-self.nh * sigma)


class Gabs(MultiplicativeComponent):
    r"""
    A Gaussian absorption model.

    $$
        \mathcal{M}(E) = \exp \left( - \frac{\tau}{\sqrt{2 \pi} \sigma} \exp
        \left( -\frac{\left(E-E_0\right)^2}{2 \sigma^2} \right) \right)
    $$

    !!! abstract "Parameters"
        * $\tau$ (`tau`) $\left[\text{dimensionless}\right]$ : Absorption strength
        * $\sigma$ (`sigma`) $\left[\text{keV}\right]$ : Absorption width
        * $E_0$ (`E0`) $\left[\text{keV}\right]$ : Absorption center

    !!! note
        The optical depth at line center is $\tau/(\sqrt{2 \pi} \sigma)$.

    """

    def __init__(self):
        self.tau = nnx.Param(1.0)
        self.sigma = nnx.Param(1.0)
        self.E0 = nnx.Param(1.0)

    def g(self, E):
        return (
            2
            / (
                self.sigma
                * jnp.sqrt(2 * jnp.pi)
                * (1 - jsp.erf(-self.E0 / (self.sigma * jnp.sqrt(2))))
            )
            * jnp.exp(-((E - self.E0) ** 2) / (2 * self.sigma**2))
        )

    def factor(self, energy):
        return jnp.exp(-self.tau * self.g(energy))


class Highecut(MultiplicativeComponent):
    r"""
    A high-energy cutoff model.

    $$
        \mathcal{M}(E) = \begin{cases} \exp
        \left( \frac{E_c - E}{E_f} \right)& \text{if $E > E_c$}\\ 1 & \text{if $E < E_c$}\end{cases}
    $$

    !!! abstract "Parameters"
        * $E_c$ (`Ec`) $\left[\text{keV}\right]$ : Cutoff energy
        * $E_f$ (`Ef`) $\left[\text{keV}\right]$ : Folding energy
    """

    def __init__(self):
        self.Ec = nnx.Param(1.0)
        self.Ef = nnx.Param(1.0)

    def factor(self, energy):
        return jnp.where(energy <= self.Ec, 1.0, jnp.exp((self.Ec - energy) / self.Ef))


class Zedge(MultiplicativeComponent):
    r"""
    A redshifted absorption edge model.

    $$
        \mathcal{M}(E) = \begin{cases} \exp \left( -D \left(\frac{E(1+z)}{E_c}\right)^3 \right)
        & \text{if $E > E_c$}\\ 1 & \text{if $E < E_c$}\end{cases}
    $$

    !!! abstract "Parameters"
        * $E_c$ (`Ec`) $\left[\text{keV}\right]$ : Threshold energy
        * $D$ (`D`) $\left[\text{dimensionless}\right]$ : Absorption depth at the threshold
        * $z$ (`z`) $\left[\text{dimensionless}\right]$ : Redshift
    """

    def __init__(self):
        self.Ec = nnx.Param(1.0)
        self.D = nnx.Param(1.0)
        self.z = nnx.Param(0.0)

    def factor(self, energy):
        return jnp.where(
            energy <= self.Ec, 1.0, jnp.exp(-self.D * (energy * (1 + self.z) / self.Ec) ** 3)
        )


class Tbpcf(MultiplicativeComponent):
    r"""
    Partial covering model using the Tuebingen-Boulder ISM absorption model (for more details, see `Tbabs`).

    $$
        \mathcal{M}(E) = f \exp^{-N_{\text{H}}\sigma(E)} + (1-f)
    $$

    !!! abstract "Parameters"
        * $N_{\text{H}}$ (`nh`) $\left[\frac{\text{atoms}~10^{22}}{\text{cm}^2}\right]$ : Equivalent hydrogen column density
        * $f$ (`f`) $\left[\text{dimensionless}\right]$ : Partial covering fraction, ranges between 0 and 1

    !!! note
        Abundances and cross-sections $\sigma$ can be found in Wilms et al. (2000).

    """

    def __init__(self):
        table = Table.read(table_manager.fetch("xsect_tbabs_wilm.fits"))
        self._energy = nnx.Variable(np.asarray(table["ENERGY"], dtype=np.float64))
        self._sigma = nnx.Variable(np.asarray(table["SIGMA"], dtype=np.float64))
        self.nh = nnx.Param(1.0)
        self.f = nnx.Param(0.2)

    def factor(self, energy):
        sigma = jnp.exp(
            jnp.interp(
                energy, self._energy, jnp.log(self._sigma), left="extrapolate", right="extrapolate"
            )
        )

        return self.f * jnp.exp(-self.nh * sigma) + (1.0 - self.f)


class FDcut(MultiplicativeComponent):
    r"""
    A Fermi-Dirac cutoff model.

    $$
        \mathcal{M}(E) = \left( 1 + \exp \left( \frac{E - E_c}{E_f} \right) \right)^{-1}
    $$

    ??? abstract "Parameters"
        * $E_c$ (`Ec`) $\left[\text{keV}\right]$ : Cutoff energy
        * $E_f$ (`Ef`) $\left[\text{keV}\right]$ : Folding energy
    """

    def __init__(self):
        self.Ec = nnx.Param(1.0)
        self.Ef = nnx.Param(3.0)

    def factor(self, energy):
        return (1 + jnp.exp((energy - self.Ec) / self.Ef)) ** -1



r"""
Module for integrating functions in a consistent way in `jaxspec`.
It mainly relies on tanh-sinh (or double exponential) quadrature to perform the integration.

!!! info "References"

    * [Takahasi and Mori (1974)](https://ems.press/journals/prims/articles/2686)
    * [Mori and Sugihara (2001)](https://doi.org/10.1016/S0377-0427(00)00501-X)
    * [Tanh-sinh quadrature](https://en.wikipedia.org/wiki/Tanh-sinh_quadrature) from Wikipedia
"""

from collections.abc import Callable

import jax
import jax.numpy as jnp

from jax import Array
from jax.scipy.integrate import trapezoid


def interval_weights(a: float, b: float, n: int) -> tuple[Array, Array, Array]:
    """
    Return the weights for the tanh-sinh quadrature over the interval [a, b].
    """
    t = jnp.linspace(-3, 3, n)
    phi = jnp.tanh(jnp.pi / 2 * jnp.sinh(t))
    dphi = jnp.pi / 2 * jnp.cosh(t) * (1 / jnp.cosh(jnp.pi / 2 * jnp.sinh(t)) ** 2)
    x = (b - a) / 2 * phi + (b + a) / 2
    dx = (b - a) / 2 * dphi

    return t, x, dx


def positive_weights(n: int) -> tuple[Array, Array, Array]:
    """
    Return the weights for the tanh-sinh quadrature over the positive real line.
    """
    t = jnp.linspace(-3, 3, n)
    x = jnp.exp(jnp.pi / 2 * jnp.sinh(t))
    dx = jnp.pi / 2 * jnp.cosh(t) * jnp.exp(jnp.pi / 2 * jnp.sinh(t))

    return t, x, dx


def integrate_interval(integrand: Callable, n: int = 51) -> Callable:
    r"""
    Build a function which can compute the integral of the provided integrand over the interval $[a, b]$ using
    the tanh-sinh quadrature. Returns a function $F(a, b, \pmb{\theta})$ which takes the limits of the interval and
    the parameters of $f(x,\pmb{\theta})$ as inputs.

    $$
    F(a, b, \pmb{\theta}) = \int_a^b f(x,\pmb{\theta}) \text{d}x
    $$

    # Example usage

    ``` python
    pi = 4*integrate_interval(lambda x: 1/(1+x**2))(0, 1)
    print(pi) # 3.1415927
    ```

    # Example where the limits of the integral are parameters
    ``` python
    def erf(x):

        def integrand(t):
            return 2/jnp.sqrt(jnp.pi) * jnp.exp(-t**2)

        return integrate_interval(integrand)(0, x)

    print(erf(1)) # 0.84270084
    ```

    Parameters:
        integrand: The function to integrate
        n: The number of points to use for the quadrature

    Returns:
        The integral of the provided integrand over the interval $[a, b]$ as a callable
    """

    @jax.custom_jvp
    def f(a, b, *args):
        t, x, dx = interval_weights(a, b, n)

        return trapezoid(jnp.nan_to_num(integrand(x, *args) * dx), x=t)

    @f.defjvp
    def f_jvp(primals, tangents):
        a, b, *args = primals
        a_dot, b_dot, *args_dot = tangents

        t, x, dx = interval_weights(a, b, n)

        primal_out = f(a, b, *args)

        # Partial derivatives along other parameters
        jac = trapezoid(
            jnp.nan_to_num(jnp.asarray(jax.jacfwd(lambda args: integrand(x, *args))(args)) * dx),
            x=t,
            axis=-1,
        )

        tangent_out = (
            -integrand(a, *args) * a_dot + integrand(b, *args) * b_dot + jac @ jnp.asarray(args_dot)
        )
        return primal_out, tangent_out

    return f


def integrate_positive(integrand: Callable, n: int = 51) -> Callable:
    r"""
    Build a function which can compute the integral of the provided integrand over the positive real line using
    the tanh-sinh quadrature. Returns a function $F(\pmb{\theta})$ which takes the parameters of the integrand
    $f(x,\pmb{\theta})$ as inputs.

    $$
    F(\pmb{\theta}) = \int_0^\infty f(x,\pmb{\theta}) \text{d}x
    $$

    # Example usage

    ``` python
    gamma = integrate_positive(lambda t, z: t**(z-1) * jnp.exp(-t))
    print(gamma(1/2)) # 1.7716383
    ```

    Parameters:
        integrand: The function to integrate
        n: The number of points to use for the quadrature

    Returns:
        The integral of the provided integrand over the positive real line as a callable
    """

    @jax.custom_jvp
    def f(*args):
        t, x, dx = positive_weights(n)

        return trapezoid(jnp.nan_to_num(integrand(x, *args) * dx), x=t)

    @f.defjvp
    def f_jvp(primals, tangents):
        args = primals
        args_dot = tangents

        t, x, dx = positive_weights(n)

        primal_out = f(*args)

        # Partial derivatives along other parameters
        jac = trapezoid(
            jnp.nan_to_num(jnp.asarray(jax.jacfwd(lambda args: integrand(x, *args))(args)) * dx),
            x=t,
            axis=-1,
        )

        tangent_out = jac @ jnp.asarray(args_dot)
        return primal_out, tangent_out

    return f



import operator

from collections.abc import Callable
from functools import cached_property
from typing import Any

import jax
import jax.numpy as jnp
import matplotlib.pyplot as plt
import numpyro

from flax import nnx
from jax.experimental import mesh_utils
from jax.random import PRNGKey
from jax.sharding import PositionalSharding
from numpyro.distributions import Poisson, TransformedDistribution
from numpyro.infer import Predictive
from numpyro.infer.inspect import get_model_relations
from numpyro.infer.reparam import TransformReparam
from numpyro.infer.util import log_density

from ..analysis._plot import (
    _error_bars_for_observed_data,
    _plot_binned_samples_with_error,
    _plot_poisson_data_with_error,
)
from ..data import ObsConfiguration
from ..model.abc import SpectralModel
from ..model.background import BackgroundModel
from ..model.instrument import InstrumentModel
from ..util.typing import PriorDictType
from ._build_model import build_prior, forward_model


class BayesianModel(nnx.Module):
    """
    Base class for a Bayesian model. This class contains the necessary methods to build a model, sample from the prior
    and compute the log-likelihood and posterior probability.
    """

    settings: dict[str, Any]

    def __init__(
        self,
        model: SpectralModel,
        prior_distributions: PriorDictType | Callable,
        observations: ObsConfiguration | list[ObsConfiguration] | dict[str, ObsConfiguration],
        background_model: BackgroundModel = None,
        instrument_model: InstrumentModel = None,
        sparsify_matrix: bool = False,
        n_points: int = 2,
    ):
        """
        Build a Bayesian model for a given spectral model and observations.

        Parameters:
            model: the spectral model to fit.
            prior_distributions: a nested dictionary containing the prior distributions for the model parameters, or a
                callable function that returns parameter samples.
            observations: the observations to fit the model to.
            background_model: the background model to fit.
            instrument_model: the instrument model to fit.
            sparsify_matrix: whether to sparsify the transfer matrix.
        """

        self.spectral_model = model
        self._observations = observations
        self.background_model = background_model
        self.instrument_model = instrument_model
        self.settings = {"sparse": sparsify_matrix}

        if not callable(prior_distributions):

            def prior_distributions_func():
                return build_prior(
                    prior_distributions,
                    expand_shape=(len(self._observation_container),),
                    prefix="mod/~/",
                )

        else:
            prior_distributions_func = prior_distributions

        self.prior_distributions_func = prior_distributions_func

        def numpyro_model(observed=True):
            # Instantiate and register the parameters of the spectral model and the background
            prior_params = self.prior_distributions_func()

            # Iterate over all the observations in our container and build a single numpyro model for each observation
            for i, (name, observation) in enumerate(self._observation_container.items()):
                # Check that we can indeed fit a background
                if (getattr(observation, "folded_background", None) is not None) and (
                    self.background_model is not None
                ):
                    # This call should register the parameter and observation of our background model
                    bkg_countrate = self.background_model.numpyro_model(
                        observation, name=name, observed=observed
                    )

                elif (getattr(observation, "folded_background", None) is None) and (
                    self.background_model is not None
                ):
                    raise ValueError(
                        "Trying to fit a background model but no background is linked to this observation"
                    )

                else:
                    bkg_countrate = 0.0

                # We expect that prior_params contains an array of parameters for each observation
                # They can be identical or different for each observation
                params = jax.tree.map(lambda x: x[i], prior_params)

                if self.instrument_model is not None:
                    gain, shift = self.instrument_model.get_gain_and_shift_model(name)
                else:
                    gain, shift = None, None

                # Forward model the observation and get the associated countrate
                obs_model = jax.jit(
                    lambda par: forward_model(
                        self.spectral_model,
                        par,
                        observation,
                        sparse=self.settings.get("sparse", False),
                        gain=gain,
                        shift=shift,
                        n_points=n_points,
                    )
                )

                obs_countrate = obs_model(params)

                # Register the observation as an observed site
                with numpyro.plate("obs_plate/~/" + name, len(observation.folded_counts)):
                    numpyro.sample(
                        "obs/~/" + name,
                        Poisson(obs_countrate + bkg_countrate / observation.folded_backratio.data),
                        obs=observation.folded_counts.data if observed else None,
                    )

        self.numpyro_model = numpyro_model
        self._init_params = self.prior_samples()
        # Check the priors are suited for the observations
        split_parameters = [
            (param, shape[-1])
            for param, shape in jax.tree.map(lambda x: x.shape, self._init_params).items()
            if (len(shape) > 1)
            and not param.startswith("_")
            and not param.startswith("bkg")  # hardcoded for subtracted background
            and not param.startswith("ins")
        ]

        for parameter, proposed_number_of_obs in split_parameters:
            if proposed_number_of_obs != len(self._observation_container):
                raise ValueError(
                    f"Invalid splitting in the prior distribution. "
                    f"Expected {len(self._observation_container)} but got {proposed_number_of_obs} for {parameter}"
                )

    @cached_property
    def _observation_container(self) -> dict[str, ObsConfiguration]:
        """
        The observations used in the fit as a dictionary of observations.
        """

        if isinstance(self._observations, dict):
            return self._observations

        elif isinstance(self._observations, list):
            return {f"data_{i}": obs for i, obs in enumerate(self._observations)}

        elif isinstance(self._observations, ObsConfiguration):
            return {"data": self._observations}

        else:
            raise ValueError(f"Invalid type for observations : {type(self._observations)}")

    @cached_property
    def transformed_numpyro_model(self) -> Callable:
        transform_dict = {}

        relations = get_model_relations(self.numpyro_model)
        distributions = {
            parameter: getattr(numpyro.distributions, value, None)
            for parameter, value in relations["sample_dist"].items()
        }

        for parameter, distribution in distributions.items():
            if isinstance(distribution, TransformedDistribution):
                transform_dict[parameter] = TransformReparam()

        return numpyro.handlers.reparam(self.numpyro_model, config=transform_dict)

    @cached_property
    def log_likelihood_per_obs(self) -> Callable:
        """
        Build the log likelihood function for each bins in each observation.
        """

        @jax.jit
        def log_likelihood_per_obs(constrained_params):
            log_likelihood = numpyro.infer.util.log_likelihood(
                model=self.numpyro_model, posterior_samples=constrained_params
            )

            return jax.tree.map(lambda x: jnp.where(jnp.isnan(x), -jnp.inf, x), log_likelihood)

        return log_likelihood_per_obs

    @cached_property
    def log_likelihood(self) -> Callable:
        """
        Build the total log likelihood function. Takes a dictionary of parameters where the keys are the parameter names
        that can be fetched with the [`parameter_names`][jaxspec.fit.BayesianModel.parameter_names].
        """

        @jax.jit
        def log_likelihood(constrained_params):
            log_likelihood = self.log_likelihood_per_obs(constrained_params)

            return jax.tree.reduce(operator.add, jax.tree.map(jnp.sum, log_likelihood))

        return log_likelihood

    @cached_property
    def log_posterior_prob(self) -> Callable:
        """
        Build the posterior probability. Takes a dictionary of parameters where the keys are the parameter names
        that can be fetched with the [`parameter_names`][jaxspec.fit.BayesianModel.parameter_names].
        """

        # This is required as numpyro.infer.util.log_densities does not check parameter validity by itself
        numpyro.enable_validation()

        @jax.jit
        def log_posterior_prob(constrained_params):
            log_posterior_prob, _ = log_density(
                self.numpyro_model, (), dict(observed=True), constrained_params
            )
            return jnp.where(jnp.isnan(log_posterior_prob), -jnp.inf, log_posterior_prob)

        return log_posterior_prob

    @cached_property
    def _parameter_names(self) -> list[str]:
        """
        A list of parameter names for the model.
        """
        relations = get_model_relations(self.numpyro_model)
        all_sites = relations["sample_sample"].keys()
        observed_sites = relations["observed"]
        return [site for site in all_sites if site not in observed_sites]

    @cached_property
    def observation_names(self) -> list[str]:
        """
        List of the observations.
        """
        relations = get_model_relations(self.numpyro_model)
        all_sites = relations["sample_sample"].keys()
        observed_sites = relations["observed"]
        return [site for site in all_sites if site in observed_sites]

    def array_to_dict(self, theta):
        """
        Convert an array of parameters to a dictionary of parameters.
        """
        input_params = {}

        for index, key in enumerate(self._parameter_names):
            input_params[key] = theta[index]

        return input_params

    def dict_to_array(self, dict_of_params):
        """
        Convert a dictionary of parameters to an array of parameters.
        """

        theta = jnp.zeros(len(self._parameter_names))

        for index, key in enumerate(self._parameter_names):
            theta = theta.at[index].set(dict_of_params[key])

        return theta

    def prior_samples(self, key: PRNGKey = PRNGKey(0), num_samples: int = 100):
        """
        Get initial parameters for the model by sampling from the prior distribution

        Parameters:
            key: the random key used to initialize the sampler.
            num_samples: the number of samples to draw from the prior.
        """

        @jax.jit
        def prior_sample(key):
            return Predictive(
                self.numpyro_model, return_sites=self._parameter_names, num_samples=num_samples
            )(key, observed=False)

        return prior_sample(key)

    def mock_observations(self, parameters, key: PRNGKey = PRNGKey(0)):
        @jax.jit
        def fakeit(key, parameters):
            return Predictive(
                self.numpyro_model,
                return_sites=self.observation_names,
                posterior_samples=parameters,
            )(key, observed=False)

        return fakeit(key, parameters)

    def prior_predictive_coverage(
        self,
        key: PRNGKey = PRNGKey(0),
        num_samples: int = 1000,
    ):
        """
        Check if the prior distribution include the observed data.
        """
        key_prior, key_posterior = jax.random.split(key, 2)
        n_devices = len(jax.local_devices())
        sharding = PositionalSharding(mesh_utils.create_device_mesh((n_devices,)))

        # Sample from prior and correct if the number of samples is not a multiple of the number of devices
        if num_samples % n_devices != 0:
            num_samples = num_samples + n_devices - (num_samples % n_devices)

        prior_params = self.prior_samples(key=key_prior, num_samples=num_samples)

        # Split the parameters on every device
        sharded_parameters = jax.device_put(prior_params, sharding)
        posterior_observations = self.mock_observations(sharded_parameters, key=key_posterior)

        for key, value in self._observation_container.items():
            fig, ax = plt.subplots(
                nrows=2, ncols=1, sharex=True, figsize=(5, 6), height_ratios=[3, 1]
            )

            legend_plots = []
            legend_labels = []

            y_observed, y_observed_low, y_observed_high = _error_bars_for_observed_data(
                value.folded_counts.values, 1.0, "ct"
            )

            true_data_plot = _plot_poisson_data_with_error(
                ax[0],
                value.out_energies,
                y_observed.value,
                y_observed_low.value,
                y_observed_high.value,
                alpha=0.7,
            )

            prior_plot = _plot_binned_samples_with_error(
                ax[0], value.out_energies, posterior_observations["obs/~/" + key], n_sigmas=3
            )

            legend_plots.append((true_data_plot,))
            legend_labels.append("Observed")
            legend_plots += prior_plot
            legend_labels.append("Prior Predictive")

            # rank = np.vstack((posterior_observations["obs_" + key], value.folded_counts.values)).argsort(axis=0)[-1] / (num_samples) * 100
            counts = posterior_observations["obs/~/" + key]
            observed = value.folded_counts.values

            num_samples = counts.shape[0]

            less_than_obs = (counts < observed).sum(axis=0)
            equal_to_obs = (counts == observed).sum(axis=0)

            rank = (less_than_obs + 0.5 * equal_to_obs) / num_samples * 100

            ax[1].stairs(rank, edges=[*list(value.out_energies[0]), value.out_energies[1][-1]])

            ax[1].plot(
                (value.out_energies.min(), value.out_energies.max()),
                (50, 50),
                color="black",
                linestyle="--",
            )

            ax[1].set_xlabel("Energy (keV)")
            ax[0].set_ylabel("Counts")
            ax[1].set_ylabel("Rank (%)")
            ax[1].set_ylim(0, 100)
            ax[0].set_xlim(value.out_energies.min(), value.out_energies.max())
            ax[0].loglog()
            ax[0].legend(legend_plots, legend_labels)
            plt.suptitle(f"Prior Predictive coverage for {key}")
            plt.tight_layout()
            plt.show()



from abc import ABC, abstractmethod

import jax
import jax.numpy as jnp
import numpyro
import numpyro.distributions as dist

from flax import nnx
from jax.scipy.integrate import trapezoid
from numpyro.distributions import Poisson
from tinygp import GaussianProcess, kernels

from .abc import SpectralModel


class BackgroundModel(ABC, nnx.Module):
    """
    Handles the background modelling in our spectra. This is handled in a separate class for now
    since backgrounds can be phenomenological models fitted directly on the folded spectrum. This is not the case for
    the source model, which is fitted on the unfolded spectrum. This might be changed later.
    """

    @abstractmethod
    def numpyro_model(self, observation, name: str = "", observed=True):
        """
        Build the model for the background.
        """
        pass


class SubtractedBackground(BackgroundModel):
    """
    Define a model where the observed background is simply subtracted from the observed.

    !!! danger

        This is not a good way to model the background, as it does not account for the fact that the measured
        background is a Poisson realisation of the true background's countrate. This is why we prefer a
        [`ConjugateBackground`][jaxspec.model.background.ConjugateBackground].

    """

    def numpyro_model(self, observation, name: str = "", observed=True):
        _, observed_counts = observation.out_energies, observation.folded_background.data
        numpyro.deterministic(f"bkg/~/{name}", observed_counts)

        return observed_counts


class BackgroundWithError(BackgroundModel):
    """
    Define a model where the observed background is subtracted from the observed accounting for its intrinsic spread. It
    fits a countrate for each background bin assuming a Poisson distribution.
    """

    def numpyro_model(self, obs, name: str = "", observed=True):
        # We can't use the build_prior_function method here because the parameter size varies
        # with the current observation. It must be instantiated in place.
        # Gamma in numpyro is parameterized by concentration and rate (alpha/beta)

        _, observed_counts = obs.out_energies, obs.folded_background.data
        alpha = observed_counts + 1
        beta = 1
        countrate = numpyro.sample(f"bkg/~/_{name}_countrate", dist.Gamma(alpha, rate=beta))

        with numpyro.plate(f"bkg/~/{name}_plate", len(observed_counts)):
            numpyro.sample(
                f"bkg/~/{name}", dist.Poisson(countrate), obs=observed_counts if observed else None
            )

        return countrate


class GaussianProcessBackground(BackgroundModel):
    """
    Define a Gaussian Process to model the background. The GP is built using the
    [`tinygp`](https://tinygp.readthedocs.io/en/stable/guide.html) library.
    """

    def __init__(
        self,
        e_min: float,
        e_max: float,
        n_nodes: int = 30,
        kernel: kernels.Kernel = kernels.Matern52,
    ):
        """
        Build the Gaussian Process background model.

        Parameters:
            e_min: The lower bound of the energy range.
            e_max: The upper bound of the energy range.
            n_nodes: The number of nodes used by the GP, must be lower than the number of channel.
            kernel: The kernel used by the GP.
        """
        self.e_min = e_min
        self.e_max = e_max
        self.n_nodes = n_nodes
        self.kernel = kernel

    def numpyro_model(self, obs, name: str = "", observed=True):
        energy, observed_counts = obs.out_energies, obs.folded_background.data

        if (observed_counts is not None) and (self.n_nodes >= len(observed_counts)):
            raise RuntimeError(
                "More nodes than channels in the observation associated with GaussianProcessBackground."
            )

        else:
            observed_counts = jnp.asarray(observed_counts)

        # The parameters of the GP model
        mean = numpyro.sample(
            f"bkg/~/_{name}_mean", dist.Normal(jnp.log(jnp.mean(observed_counts)), 2.0)
        )
        sigma = numpyro.sample(f"bkg/~/_{name}_sigma", dist.HalfNormal(3.0))
        rho = numpyro.sample(f"bkg/~/_{name}_rho", dist.HalfNormal(10.0))

        # Set up the kernel and GP objects
        kernel = sigma**2 * self.kernel(rho)
        nodes = jnp.linspace(0, 1, self.n_nodes)
        gp = GaussianProcess(kernel, nodes, diag=1e-5 * jnp.ones_like(nodes), mean=mean)

        log_rate = numpyro.sample(f"bkg/~/_{name}_log_rate_nodes", gp.numpyro_dist())

        interp_count_rate = jnp.exp(
            jnp.interp(energy, nodes * (self.e_max - self.e_min) + self.e_min, log_rate)
        )

        count_rate = trapezoid(interp_count_rate, energy, axis=0)

        # Finally, our observation model is Poisson
        with numpyro.plate("bkg/~/plate_" + name, len(observed_counts)):
            numpyro.sample(
                f"bkg/~/{name}", dist.Poisson(count_rate), obs=observed_counts if observed else None
            )

        return count_rate


class SpectralModelBackground(BackgroundModel):
    def __init__(self, spectral_model: "SpectralModel", prior_distributions, sparse=False):
        self.spectral_model = spectral_model
        self.prior = prior_distributions
        self.sparse = sparse

    def numpyro_model(self, observation, name: str = "", observed=True):
        from jaxspec.fit._build_model import build_prior, forward_model

        params = build_prior(self.prior, prefix=f"_bkg_{name}_")
        bkg_model = jax.jit(
            lambda par: forward_model(self.spectral_model, par, observation, sparse=self.sparse)
        )
        bkg_countrate = bkg_model(params)

        with numpyro.plate("bkg/~/plate_" + name, len(observation.folded_background)):
            numpyro.sample(
                "bkg/~/" + name,
                Poisson(bkg_countrate),
                obs=observation.folded_background.data if observed else None,
            )

        return bkg_countrate


'''
class ConjugateBackground(BackgroundModel):
    r"""
    This class fit an expected rate $\\lambda$ in each bin of the background spectrum. Assuming a Gamma prior
    distribution, we can analytically derive the posterior as a Negative binomial distribution.

    $$ p(\\lambda_{\text{Bkg}}) \\sim \\Gamma \\left( \alpha, \beta \right) \\implies
    p\\left(\\lambda_{\text{Bkg}} | \text{Counts}_{\text{Bkg}}\right) \\sim \text{NB}\\left(\alpha, \frac{\beta}{\beta +1}
    \right) $$

    !!! info
        Here, $\alpha$ and $\beta$ are set to $\alpha = \text{Counts}_{\text{Bkg}} + 1$ and $\beta = 1$. Doing so,
        the prior distribution is such that $\\mathbb{E}[\\lambda_{\text{Bkg}}] = \text{Counts}_{\text{Bkg}} +1$ and
        $\text{Var}[\\lambda_{\text{Bkg}}] = \text{Counts}_{\text{Bkg}}+1$. The +1 is to avoid numerical issues when the
        counts are 0, and add a small scatter even if the measured background is effectively null.

    ??? abstract "References"

        - https://en.wikipedia.org/wiki/Conjugate_prior
        - https://www.acsu.buffalo.edu/~adamcunn/probability/gamma.html
        - https://bayesiancomputationbook.com/markdown/chp_01.html?highlight=conjugate#conjugate-priors
        - https://vioshyvo.github.io/Bayesian_inference/conjugate-distributions.html

    """

    def numpyro_model(self, energy, observed_counts, name: str = "bkg", observed=True):
        # Gamma in numpyro is parameterized by concentration and rate (alpha/beta)
        # alpha = observed_counts + 1
        # beta = 1

        with numpyro.plate(f"{name}_plate", len(observed_counts)):
            countrate = numpyro.sample(f"{name}", dist.Gamma(2 * observed_counts + 1, 2), obs=None)

        return countrate
'''



import operator

import jax
import jax.numpy as jnp

from jaxspec.fit import BayesianModel


def test_likelihood(obs_model_prior):
    obsconf, model, prior = obs_model_prior
    bayesian_model = BayesianModel(model, prior, obsconf)

    parameter_array = jnp.asarray([[0.7, 0.2, 2, 3e-4]])
    parameters = bayesian_model.array_to_dict(parameter_array)

    total_likelihood = bayesian_model.log_likelihood(parameters)
    splitted_likelihood = bayesian_model.log_likelihood_per_obs(parameters)
    total_likelihood_from_splitted = jax.tree.reduce(
        operator.add, jax.tree.map(jnp.sum, splitted_likelihood)
    )

    assert jnp.isclose(total_likelihood_from_splitted, total_likelihood)



import os

import astropy.units as u
import numpy as np
import sparse

from astropy.io import fits
from astropy.table import QTable


class DataPHA:
    r"""
    Class to handle PHA data defined with OGIP standards.
    ??? info "References"
        * [The OGIP standard PHA file format](https://heasarc.gsfc.nasa.gov/docs/heasarc/ofwg/docs/spectra/ogip_92_007/node5.html)
    """

    def __init__(
        self,
        channel,
        counts,
        exposure,
        grouping=None,
        quality=None,
        backfile=None,
        respfile=None,
        ancrfile=None,
        backscal=1.0,
        areascal=1.0,
        flags=None,
    ):
        self.channel = np.asarray(channel, dtype=int)
        self.counts = np.asarray(counts, dtype=int)
        self.exposure = float(exposure)

        self.quality = np.asarray(quality, dtype=int)
        self.backfile = backfile
        self.respfile = respfile
        self.ancrfile = ancrfile
        self.backscal = np.asarray(backscal, dtype=float)
        self.areascal = np.asarray(areascal, dtype=float)
        self.flags = flags

        if grouping is not None:
            # Indices array of the beginning of each group
            b_grp = np.where(grouping == 1)[0]
            # Indices array of the ending of each group
            e_grp = np.hstack((b_grp[1:], len(channel)))

            # Prepare data for sparse matrix
            rows = []
            cols = []
            data = []

            for i in range(len(b_grp)):
                for j in range(b_grp[i], e_grp[i]):
                    rows.append(i)
                    cols.append(j)
                    data.append(True)

            # Create a COO sparse matrix
            grp_matrix = sparse.COO(
                (data, (rows, cols)), shape=(len(b_grp), len(channel)), fill_value=0
            )

        else:
            # Identity matrix case, use sparse for efficiency
            grp_matrix = sparse.eye(len(channel), format="coo", dtype=bool)

        self.grouping = grp_matrix

    @classmethod
    def from_file(cls, pha_file: str | os.PathLike):
        """
        Load the data from a PHA file.

        Parameters:
            pha_file: The PHA file path.
        """

        data = QTable.read(pha_file, "SPECTRUM")
        header = fits.getheader(pha_file, "SPECTRUM")
        flags = []

        if header.get("HDUCLAS3") == "RATE":
            raise ValueError(
                f"The HDUCLAS3={header.get('HDUCLAS3')} keyword in the PHA file is not supported."
                f"Please open an issue if this is required."
            )

        if header.get("HDUCLAS4") == "TYPE:II":
            raise ValueError(
                f"The HDUCLAS4={header.get('HDUCLAS4')} keyword in the PHA file is not supported."
                f"Please open an issue if this is required."
            )

        if header.get("GROUPING") == 0:
            grouping = None
        elif "GROUPING" in data.colnames:
            grouping = data["GROUPING"]
        else:
            raise ValueError("No grouping column found in the PHA file.")

        if header.get("QUALITY") == 0:
            quality = np.zeros(len(data["CHANNEL"]), dtype=bool)
        elif "QUALITY" in data.colnames:
            quality = data["QUALITY"]
        else:
            raise ValueError("No QUALITY column found in the PHA file.")

        if "BACKSCAL" in header:
            backscal = header["BACKSCAL"] * np.ones_like(data["CHANNEL"], dtype=float)
        elif "BACKSCAL" in data.colnames:
            backscal = data["BACKSCAL"]
        else:
            raise ValueError("No BACKSCAL found in the PHA file.")

        backscal = np.where(backscal == 0, 1.0, backscal)

        if "AREASCAL" in header:
            areascal = header["AREASCAL"]
        elif "AREASCAL" in data.colnames:
            areascal = data["AREASCAL"]
        else:
            raise ValueError("No AREASCAL found in the PHA file.")

        if header.get("HDUCLAS2") == "NET":
            flags.append("NET")

        kwargs = {
            "grouping": grouping,
            "quality": quality,
            "backfile": header.get("BACKFILE"),
            "respfile": header.get("RESPFILE"),
            "ancrfile": header.get("ANCRFILE"),
            "backscal": backscal,
            "areascal": areascal,
            "flags": flags,
        }

        if "COUNTS" in data.colnames:
            counts = data["COUNTS"]
        elif "RATE" in data.colnames:
            counts = data["RATE"] * header["EXPOSURE"]
        else:
            raise ValueError("No COUNTS or RATE column found in the PHA file.")

        return cls(data["CHANNEL"], counts, header["EXPOSURE"], **kwargs)


class DataARF:
    r"""
    Class to handle ARF data defined with OGIP standards.

    ??? info "References"
        * [The Calibration Requirements for Spectral Analysis (Definition of RMF and ARF file formats)](https://heasarc.gsfc.nasa.gov/docs/heasarc/caldb/docs/memos/cal_gen_92_002/cal_gen_92_002.html)
        * [The Calibration Requirements for Spectral Analysis Addendum: Changes log](https://heasarc.gsfc.nasa.gov/docs/heasarc/caldb/docs/memos/cal_gen_92_002a/cal_gen_92_002a.html)
    """

    def __init__(self, energ_lo, energ_hi, specresp):
        self.specresp = specresp
        self.energ_lo = energ_lo
        self.energ_hi = energ_hi

    @classmethod
    def from_file(cls, arf_file: str | os.PathLike):
        """
        Load the data from an ARF file.

        Parameters:
            arf_file: The ARF file path.
        """

        arf_table = QTable.read(arf_file)

        return cls(
            arf_table["ENERG_LO"],
            arf_table["ENERG_HI"],
            arf_table["SPECRESP"].to(u.cm**2).value,
        )


class DataRMF:
    r"""
    Class to handle RMF data defined with OGIP standards.
    ??? info "References"
        * [The Calibration Requirements for Spectral Analysis (Definition of RMF and ARF file formats)](https://heasarc.gsfc.nasa.gov/docs/heasarc/caldb/docs/memos/cal_gen_92_002/cal_gen_92_002.html)
        * [The Calibration Requirements for Spectral Analysis Addendum: Changes log](https://heasarc.gsfc.nasa.gov/docs/heasarc/caldb/docs/memos/cal_gen_92_002a/cal_gen_92_002a.html)
    """

    def __init__(
        self,
        energ_lo,
        energ_hi,
        n_grp,
        f_chan,
        n_chan,
        matrix,
        channel,
        e_min,
        e_max,
        low_threshold=0.0,
    ):
        # RMF stuff
        self.energ_lo = energ_lo  # "Entry" energies
        self.energ_hi = energ_hi  # "Entry" energies
        self.n_grp = n_grp
        self.f_chan = f_chan
        self.n_chan = n_chan
        self.matrix_entry = matrix

        # Detector channels
        self.channel = channel
        self.e_min = e_min
        self.e_max = e_max

        # Prepare data for sparse matrix
        rows = []
        cols = []
        data = []

        for i, n_grp_val in enumerate(self.n_grp):
            base = 0

            if np.size(self.f_chan[i]) == 1:
                low = int(self.f_chan[i].ravel()[0])
                high = min(
                    int(self.f_chan[i].ravel()[0] + self.n_chan[i].ravel()[0]),
                    len(self.channel),
                )

                rows.extend([i] * (high - low))
                cols.extend(range(low, high))
                data.extend(self.matrix_entry[i][0 : high - low])

            else:
                for j in range(n_grp_val):
                    low = self.f_chan[i][j]
                    high = min(self.f_chan[i][j] + self.n_chan[i][j], len(self.channel))

                    rows.extend([i] * (high - low))
                    cols.extend(range(low, high))
                    data.extend(self.matrix_entry[i][base : base + self.n_chan[i][j]])

                    base += self.n_chan[i][j]

        # Convert lists to numpy arrays
        rows = np.array(rows)
        cols = np.array(cols)
        data = np.array(data)

        # Sometimes, zero elements are given in the matrix rows, so we get rid of them
        idxs = data > low_threshold

        # Create a COO sparse matrix and then convert to CSR for efficiency
        coo = sparse.COO(
            [rows[idxs], cols[idxs]], data[idxs], shape=(len(self.energ_lo), len(self.channel))
        )
        self.sparse_matrix = coo.T  # .tocsr()

    @property
    def matrix(self):
        return np.asarray(self.sparse_matrix.todense())

    @classmethod
    def from_file(cls, rmf_file: str | os.PathLike):
        """
        Load the data from an RMF file.

        Parameters:
            rmf_file: The RMF file path.
        """
        extension_names = [hdu[1] for hdu in fits.info(rmf_file, output=False)]

        if "MATRIX" in extension_names:
            matrix_extension = "MATRIX"

        elif "SPECRESP MATRIX" in extension_names:
            matrix_extension = "SPECRESP MATRIX"
            # raise NotImplementedError("SPECRESP MATRIX extension is not yet supported")

        else:
            raise ValueError("No MATRIX or SPECRESP MATRIX extension found in the RMF file")

        matrix_table = QTable.read(rmf_file, matrix_extension)
        ebounds_table = QTable.read(rmf_file, "EBOUNDS")

        matrix_header = fits.getheader(rmf_file, matrix_extension)

        f_chan_column_pos = list(matrix_table.columns).index("F_CHAN") + 1
        tlmin_fchan = int(matrix_header[f"TLMIN{f_chan_column_pos}"])

        return cls(
            matrix_table["ENERG_LO"],
            matrix_table["ENERG_HI"],
            matrix_table["N_GRP"],
            matrix_table["F_CHAN"] - tlmin_fchan,
            matrix_table["N_CHAN"],
            matrix_table["MATRIX"],
            ebounds_table["CHANNEL"],
            ebounds_table["E_MIN"],
            ebounds_table["E_MAX"],
        )



from jaxspec.fit import MCMCFitter


def test_sparsify_matrix_in_model(obs_model_prior):
    obsconfigurations, model, prior = obs_model_prior

    for obsconf in obsconfigurations:
        forward_model = MCMCFitter(
            model, prior, obsconf, background_model=None, sparsify_matrix=True
        )
        forward_model.fit(
            num_chains=4, num_warmup=10, num_samples=10, mcmc_kwargs={"progress_bar": False}
        )
